import logging
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, Union
from uuid import uuid4

import pandas as pd
import pyarrow as pa
import pyspark.sql as ps
from py4j.protocol import Py4JError
from pyspark import StorageLevel
from pyspark.rdd import RDD
from pyspark.sql import SparkSession
from pyspark.sql.functions import broadcast, col, lit, row_number
from pyspark.sql.window import Window
from triad import IndexedOrderedDict, ParamDict, Schema, SerializableRLock
from triad.utils.assertion import assert_arg_not_none, assert_or_throw
from triad.utils.hash import to_uuid
from triad.utils.iter import EmptyAwareIterable
from triad.utils.pandas_like import PD_UTILS
from triad.utils.pyarrow import get_alter_func
from triad.utils.threading import RunOnce

from fugue import StructuredRawSQL
from fugue.collections.partition import (
    PartitionCursor,
    PartitionSpec,
    parse_presort_exp,
)
from fugue.constants import KEYWORD_PARALLELISM, KEYWORD_ROWCOUNT
from fugue.dataframe import (
    ArrayDataFrame,
    ArrowDataFrame,
    DataFrame,
    DataFrames,
    IterableArrowDataFrame,
    IterableDataFrame,
    IterablePandasDataFrame,
    LocalDataFrame,
    LocalDataFrameIterableDataFrame,
    PandasDataFrame,
)
from fugue.dataframe.utils import get_join_schemas
from fugue.exceptions import FugueDataFrameInitError
from fugue.execution.execution_engine import ExecutionEngine, MapEngine, SQLEngine

from ._constants import FUGUE_SPARK_CONF_USE_PANDAS_UDF, FUGUE_SPARK_DEFAULT_CONF
from ._utils.convert import (
    _PYSPARK_ARROW_FRIENDLY,
    pandas_udf_can_accept,
    to_schema,
    to_spark_df,
    to_spark_schema,
    to_type_safe_input,
)
from ._utils.io import SparkIO
from ._utils.misc import is_spark_connect as _is_spark_connect
from ._utils.misc import is_spark_dataframe
from ._utils.partition import even_repartition, hash_repartition, rand_repartition
from .dataframe import SparkDataFrame

_TO_SPARK_JOIN_MAP: Dict[str, str] = {
    "inner": "inner",
    "leftouter": "left_outer",
    "rightouter": "right_outer",
    "fullouter": "outer",
    "cross": "cross",
    "semi": "left_semi",
    "anti": "left_anti",
    "leftsemi": "left_semi",
    "leftanti": "left_anti",
}


class SparkSQLEngine(SQLEngine):
    """`Spark SQL <https://spark.apache.org/sql/>`_ execution implementation.

    :param execution_engine: it must be :class:`~.SparkExecutionEngine`
    :raises ValueError: if the engine is not :class:`~.SparkExecutionEngine`
    """

    @property
    def dialect(self) -> Optional[str]:
        return "spark"

    @property
    def execution_engine_constraint(self) -> Type[ExecutionEngine]:
        return SparkExecutionEngine

    @property
    def is_distributed(self) -> bool:
        return True

    def select(self, dfs: DataFrames, statement: StructuredRawSQL) -> DataFrame:
        _map: Dict[str, str] = {}
        for k, v in dfs.items():
            df = self.execution_engine._to_spark_df(v, create_view=True)  # type: ignore
            _map[k] = df.alias
        _sql = statement.construct(_map, dialect=self.dialect, log=self.log)
        return SparkDataFrame(
            self.execution_engine.spark_session.sql(_sql)  # type: ignore
        )


class SparkMapEngine(MapEngine):
    @property
    def is_distributed(self) -> bool:
        return True

    @property
    def is_spark_connect(self) -> bool:
        """Whether the spark session is created by spark connect"""
        return self.execution_engine.is_spark_connect  # type:ignore

    def _should_use_pandas_udf(
        self, input_schema: Schema, output_schema: Schema
    ) -> bool:
        if self.is_spark_connect:  # pragma: no cover
            return True
        # else:  # this condition seems to be unnecessary
        #    possible &= self.execution_engine.conf.get(
        #        "spark.sql.execution.arrow.pyspark.enabled", False
        #    )
        compatible = pandas_udf_can_accept(
            input_schema, is_input=True
        ) and pandas_udf_can_accept(output_schema, is_input=False)
        enabled = self.execution_engine.conf.get_or_throw(
            FUGUE_SPARK_CONF_USE_PANDAS_UDF, bool
        )
        if enabled and not compatible:  # pragma: no cover
            self.log.warning(
                f"{FUGUE_SPARK_CONF_USE_PANDAS_UDF}"
                f" is enabled but {input_schema} or {output_schema}"
                " is not compatible with pandas udf, using RDD instead"
            )
            return False
        return enabled

    def map_dataframe(
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
        map_func_format_hint: Optional[str] = None,
    ) -> DataFrame:
        output_schema = Schema(output_schema)
        if self._should_use_pandas_udf(df.schema, output_schema):
            if len(partition_spec.partition_by) > 0:
                if partition_spec.algo in ["coarse", "even"]:
                    return self._map_by_pandas_udf(
                        df,
                        map_func=map_func,
                        output_schema=output_schema,
                        partition_spec=partition_spec,
                        on_init=on_init,
                        map_func_format_hint=map_func_format_hint,
                    )
                else:
                    if (  # not simple partitioning
                        partition_spec.algo not in ["default", "hash"]
                        or partition_spec.num_partitions != "0"
                    ):
                        # TODO: not sure if presort should be done
                        # on physical partition level
                        df = self.to_df(
                            self.execution_engine.repartition(
                                df, PartitionSpec(partition_spec, presort=[])
                            )
                        )
                    return self._group_map_by_pandas_udf(
                        df,
                        map_func=map_func,
                        output_schema=output_schema,
                        partition_spec=partition_spec,
                        on_init=on_init,
                        map_func_format_hint=map_func_format_hint,
                    )
            else:
                return self._map_by_pandas_udf(
                    df,
                    map_func=map_func,
                    output_schema=output_schema,
                    partition_spec=partition_spec,
                    on_init=on_init,
                    map_func_format_hint=map_func_format_hint,
                )
        df = self.to_df(self.execution_engine.repartition(df, partition_spec))
        mapper = _Mapper(df, map_func, output_schema, partition_spec, on_init)
        sdf = df.native.rdd.mapPartitionsWithIndex(mapper.run, True)  # type: ignore
        return self.to_df(sdf, output_schema)

    def _group_map_by_pandas_udf(
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
        map_func_format_hint: Optional[str] = None,
    ) -> DataFrame:
        presort = partition_spec.presort
        presort_keys = list(presort.keys())
        presort_asc = list(presort.values())
        output_schema = Schema(output_schema)
        input_schema = df.schema
        cursor = partition_spec.get_cursor(input_schema, 0)
        on_init_once: Any = (
            None
            if on_init is None
            else RunOnce(
                on_init, lambda *args, **kwargs: to_uuid(id(on_init), id(args[0]))
            )
        )

        def _udf_pandas(pdf: Any) -> pd.DataFrame:  # pragma: no cover
            if pdf.shape[0] == 0:
                return _to_safe_spark_worker_pandas(
                    PandasDataFrame(schema=output_schema).as_pandas()
                )
            if len(partition_spec.presort) > 0:
                pdf = pdf.sort_values(presort_keys, ascending=presort_asc)
            input_df = PandasDataFrame(
                pdf.reset_index(drop=True), input_schema, pandas_df_wrapper=True
            )
            if on_init_once is not None:
                on_init_once(0, input_df)
            cursor.set(lambda: input_df.peek_array(), 0, 0)
            output_df = map_func(cursor, input_df)
            return _to_safe_spark_worker_pandas(output_df.as_pandas())

        df = self.to_df(df)

        gdf = df.native.groupBy(*partition_spec.partition_by)  # type: ignore
        sdf = gdf.applyInPandas(_udf_pandas, schema=to_spark_schema(output_schema))
        return SparkDataFrame(sdf)

    def _map_by_pandas_udf(  # noqa: C901
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
        map_func_format_hint: Optional[str] = None,
    ) -> DataFrame:
        df = self.to_df(self.execution_engine.repartition(df, partition_spec))
        output_schema = Schema(output_schema)
        input_schema = df.schema
        cursor = partition_spec.get_cursor(input_schema, 0)
        on_init_once: Any = (
            None
            if on_init is None
            else RunOnce(
                on_init, lambda *args, **kwargs: to_uuid(id(on_init), id(args[0]))
            )
        )

        def _udf_pandas(
            dfs: Iterable[pd.DataFrame],
        ) -> Iterable[pd.DataFrame]:  # pragma: no cover
            def get_dfs() -> Iterable[LocalDataFrame]:
                cursor_set = False
                for df in dfs:
                    if df.shape[0] > 0:
                        pdf = PandasDataFrame(
                            df.reset_index(drop=True),
                            input_schema,
                            pandas_df_wrapper=True,
                        )
                        if not cursor_set:
                            cursor.set(lambda: pdf.peek_array(), 0, 0)
                            cursor_set = True
                        yield pdf

            input_df = IterablePandasDataFrame(get_dfs(), input_schema)
            if input_df.empty:
                yield _to_safe_spark_worker_pandas(
                    PandasDataFrame(schema=output_schema).as_pandas()
                )
                return
            if on_init_once is not None:
                on_init_once(0, input_df)
            output_df = map_func(cursor, input_df)
            if isinstance(output_df, LocalDataFrameIterableDataFrame):
                for res in output_df.native:
                    yield _to_safe_spark_worker_pandas(res.as_pandas())
            else:
                yield _to_safe_spark_worker_pandas(output_df.as_pandas())

        def _udf_arrow(
            dfs: Iterable[pa.RecordBatch],
        ) -> Iterable[pa.RecordBatch]:  # pragma: no cover
            def get_dfs() -> Iterable[LocalDataFrame]:
                cursor_set = False
                func: Any = None
                for df in dfs:
                    if df.num_rows > 0:
                        # TODO: need coalesce based on byte size
                        adf = pa.Table.from_batches([df])
                        if func is None:
                            # mapInArrow has bugs, the input timestamp will
                            # be added with a timezong, this is to fix the issue
                            func = get_alter_func(
                                adf.schema, input_schema.pa_schema, safe=False
                            )
                        pdf = ArrowDataFrame(func(adf))
                        if not cursor_set:
                            cursor.set(lambda: pdf.peek_array(), 0, 0)
                            cursor_set = True
                        yield pdf

            input_df = IterableArrowDataFrame(get_dfs(), input_schema)
            if input_df.empty:
                yield from output_schema.create_empty_arrow_table().to_batches()
                return
            if on_init_once is not None:
                on_init_once(0, input_df)
            output_df = map_func(cursor, input_df)
            if isinstance(output_df, LocalDataFrameIterableDataFrame):
                for res in output_df.native:
                    yield from res.as_arrow().to_batches()
            else:
                yield from output_df.as_arrow().to_batches()

        df = self.to_df(df)
        if map_func_format_hint == "pyarrow" and hasattr(df.native, "mapInArrow"):
            sdf = df.native.mapInArrow(  # type: ignore
                _udf_arrow, schema=to_spark_schema(output_schema)
            )
        else:
            sdf = df.native.mapInPandas(  # type: ignore
                _udf_pandas, schema=to_spark_schema(output_schema)
            )
        return SparkDataFrame(sdf)


class SparkExecutionEngine(ExecutionEngine):
    """The execution engine based on :class:`~spark:pyspark.sql.SparkSession`.

    Please read |ExecutionEngineTutorial| to understand this important Fugue concept

    :param spark_session: Spark session, defaults to None to get the Spark session by
      :meth:`~spark:pyspark.sql.SparkSession.Builder.getOrCreate`
    :param conf: |ParamsLikeObject| defaults to None, read |FugueConfig| to
      learn Fugue specific options
    """

    def __init__(self, spark_session: Optional[SparkSession] = None, conf: Any = None):
        if spark_session is None:
            spark_session = SparkSession.builder.getOrCreate()
        self._spark_session = spark_session
        cf = dict(FUGUE_SPARK_DEFAULT_CONF)
        if not self.is_spark_connect:
            try:
                spark_conf = spark_session.sparkContext.getConf()
                cf.update({x[0]: x[1] for x in spark_conf.getAll()})
            except Py4JError:  # pragma: no cover:
                # edge case: https://github.com/fugue-project/fugue/issues/517z
                pass
        cf.update(ParamDict(conf))
        super().__init__(cf)
        self._lock = SerializableRLock()
        self._log = logging.getLogger()
        self._broadcast_func = RunOnce(
            self._broadcast, lambda *args, **kwargs: id(args[0])
        )
        self._persist_func = RunOnce(self._persist, lambda *args, **kwargs: id(args[0]))
        self._io = SparkIO(self.spark_session)
        self._registered_dfs: Dict[str, SparkDataFrame] = {}

    def __repr__(self) -> str:
        return "SparkExecutionEngine"

    @property
    def spark_session(self) -> SparkSession:
        """
        :return: The wrapped spark session
        :rtype: :class:`spark:pyspark.sql.SparkSession`
        """
        assert_or_throw(
            self._spark_session is not None, "SparkExecutionEngine is not started"
        )
        return self._spark_session

    @property
    def is_spark_connect(self) -> bool:
        return _is_spark_connect(self.spark_session)

    @property
    def is_distributed(self) -> bool:
        return True

    @property
    def log(self) -> logging.Logger:
        return self._log

    def create_default_sql_engine(self) -> SQLEngine:
        return SparkSQLEngine(self)

    def create_default_map_engine(self) -> MapEngine:
        return SparkMapEngine(self)

    def get_current_parallelism(self) -> int:
        spark = self.spark_session
        if self.is_spark_connect:  # pragma: no cover
            num = spark.conf.get("spark.default.parallelism", "")
            if num != "":
                return int(num)
            return int(spark.conf.get("spark.sql.shuffle.partitions", "200"))
        e_cores = int(spark.conf.get("spark.executor.cores", "1"))
        tc = int(spark.conf.get("spark.task.cpus", "1"))
        sc = spark._jsc.sc()
        try:
            return spark.sparkContext.defaultParallelism // tc
        except Exception:  # pragma: no cover
            # for pyspark < 3.1.1
            nodes = len(list(sc.statusTracker().getExecutorInfos()))
            workers = 1 if nodes <= 1 else nodes - 1
            return max(workers * (e_cores // tc), 1)

    def to_df(self, df: Any, schema: Any = None) -> SparkDataFrame:  # noqa: C901
        """Convert a data structure to :class:`~fugue_spark.dataframe.SparkDataFrame`

        :param data: :class:`~fugue.dataframe.dataframe.DataFrame`,
          :class:`spark:pyspark.sql.DataFrame`, :class:`spark:pyspark.RDD`,
          pandas DataFrame or list or iterable of arrays
        :param schema: |SchemaLikeObject| or :class:`spark:pyspark.sql.types.StructType`
          defaults to None.
        :return: engine compatible dataframe

        .. note::

            * if the input is already :class:`~fugue_spark.dataframe.SparkDataFrame`,
              it should return itself
            * For :class:`~spark:pyspark.RDD`, list or iterable of arrays,
              ``schema`` must be specified
            * When ``schema`` is not None, a potential type cast may happen to ensure
              the dataframe's schema.
            * all other methods in the engine can take arbitrary dataframes and
              call this method to convert before doing anything
        """
        return self._to_spark_df(df, schema=schema)

    def repartition(self, df: DataFrame, partition_spec: PartitionSpec) -> DataFrame:
        def _persist_and_count(df: DataFrame) -> int:
            df = self.persist(df)
            return df.count()

        df = self._to_spark_df(df)
        num_funcs = {
            KEYWORD_ROWCOUNT: lambda: _persist_and_count(df),
            KEYWORD_PARALLELISM: lambda: self.get_current_parallelism(),
        }
        num = partition_spec.get_num_partitions(**num_funcs)

        if partition_spec.algo in ["default", "hash", "coarse"]:
            sdf = hash_repartition(
                self.spark_session, df.native, num, partition_spec.partition_by
            )
        elif partition_spec.algo == "rand":
            sdf = rand_repartition(
                self.spark_session, df.native, num, partition_spec.partition_by
            )
        elif partition_spec.algo == "even":
            df = self.persist(df)
            sdf = even_repartition(
                self.spark_session, df.native, num, partition_spec.partition_by
            )
        else:  # pragma: no cover
            raise NotImplementedError(partition_spec.algo + " is not supported")
        sorts = partition_spec.get_sorts(df.schema)
        if len(sorts) > 0:
            sdf = sdf.sortWithinPartitions(
                *sorts.keys(), ascending=list(sorts.values())
            )
        return self._to_spark_df(sdf, df.schema)

    def broadcast(self, df: DataFrame) -> SparkDataFrame:
        res = self._broadcast_func(self._to_spark_df(df))
        res.reset_metadata(df.metadata)
        return res

    def persist(
        self,
        df: DataFrame,
        lazy: bool = False,
        **kwargs: Any,
    ) -> SparkDataFrame:
        res = self._persist_func(
            self._to_spark_df(df), lazy=lazy, level=kwargs.get("level", None)
        )
        res.reset_metadata(df.metadata)
        return res

    def register(self, df: DataFrame, name: str) -> SparkDataFrame:
        sdf = self._to_spark_df(df)
        sdf.native.createOrReplaceTempView(name)
        return sdf

    def join(
        self,
        df1: DataFrame,
        df2: DataFrame,
        how: str,
        on: Optional[List[str]] = None,
    ) -> DataFrame:
        key_schema, output_schema = get_join_schemas(df1, df2, how=how, on=on)
        how = how.lower().replace("_", "").replace(" ", "")
        assert_or_throw(
            how in _TO_SPARK_JOIN_MAP,
            ValueError(f"{how} is not supported as a join type"),
        )
        how = _TO_SPARK_JOIN_MAP[how]
        d1 = self._to_spark_df(df1).native
        d2 = self._to_spark_df(df2).native
        cols = [col(n) for n in output_schema.names]
        if how == "cross":
            res = d1.crossJoin(d2).select(*cols)
        else:
            res = d1.join(d2, on=key_schema.names, how=how).select(*cols)
        return self._to_spark_df(res, output_schema)

    def union(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d1 = self._to_spark_df(df1).native
        d2 = self._to_spark_df(df2).native
        d = d1.union(d2)
        if distinct:
            d = d.distinct()
        return self._to_spark_df(d, df1.schema)

    def subtract(
        self, df1: DataFrame, df2: DataFrame, distinct: bool = True
    ) -> DataFrame:
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d1 = self._to_spark_df(df1).native
        d2 = self._to_spark_df(df2).native
        if distinct:
            d: Any = d1.subtract(d2)
        else:  # pragma: no cover
            d = d1.exceptAll(d2)
        return self._to_spark_df(d, df1.schema)

    def intersect(
        self, df1: DataFrame, df2: DataFrame, distinct: bool = True
    ) -> DataFrame:
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d1 = self._to_spark_df(df1).native
        d2 = self._to_spark_df(df2).native
        if distinct:
            d: Any = d1.intersect(d2)
        else:  # pragma: no cover
            d = d1.intersectAll(d2)
        return self._to_spark_df(d, df1.schema)

    def distinct(self, df: DataFrame) -> DataFrame:
        d = self._to_spark_df(df).native.distinct()
        return self._to_spark_df(d, df.schema)

    def dropna(
        self,
        df: DataFrame,
        how: str = "any",
        thresh: int = None,
        subset: List[str] = None,
    ) -> DataFrame:
        d = self._to_spark_df(df).native.dropna(how=how, thresh=thresh, subset=subset)
        return self._to_spark_df(d, df.schema)

    def fillna(self, df: DataFrame, value: Any, subset: List[str] = None) -> DataFrame:
        assert_or_throw(
            (not isinstance(value, list)) and (value is not None),
            ValueError("fillna value can not be a list or None"),
        )
        if isinstance(value, dict):
            assert_or_throw(
                (None not in value.values()) and (any(value.values())),
                ValueError(
                    "fillna dict can not contain None and needs at least one value"
                ),
            )
            mapping = value
        else:
            # If subset is none, apply to all columns
            subset = subset or df.columns
            mapping = {col: value for col in subset}
        d = self._to_spark_df(df).native.fillna(mapping)
        return self._to_spark_df(d, df.schema)

    def sample(
        self,
        df: DataFrame,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        replace: bool = False,
        seed: Optional[int] = None,
    ) -> DataFrame:
        assert_or_throw(
            (n is None and frac is not None) or (n is not None and frac is None),
            ValueError("one and only one of n and frac should be set"),
        )
        if frac is not None:
            d = self._to_spark_df(df).native.sample(
                fraction=frac, withReplacement=replace, seed=seed
            )
            return self._to_spark_df(d, df.schema)
        else:
            assert_or_throw(
                seed is None,
                NotImplementedError("seed can't be set when sampling by row count"),
            )
            assert_or_throw(
                not replace,
                NotImplementedError(
                    "replacement is not supported when sampling by row count"
                ),
            )
            temp_name = "__temp_" + str(uuid4()).split("-")[-1]
            self._to_spark_df(df).native.createOrReplaceTempView(temp_name)
            d = self.spark_session.sql(
                f"SELECT * FROM {temp_name} TABLESAMPLE({n} ROWS)"
            )
            return self._to_spark_df(d, df.schema)

    def take(
        self,
        df: DataFrame,
        n: int,
        presort: str,
        na_position: str = "last",
        partition_spec: Optional[PartitionSpec] = None,
    ) -> DataFrame:
        partition_spec = partition_spec or PartitionSpec()
        assert_or_throw(
            isinstance(n, int),
            ValueError("n needs to be an integer"),
        )
        d = self._to_spark_df(df).native
        nulls_last = bool(na_position == "last")

        if presort:
            presort = parse_presort_exp(presort)
        # Use presort over partition_spec.presort if possible
        _presort: IndexedOrderedDict = presort or partition_spec.presort

        def _presort_to_col(_col: str, _asc: bool) -> Any:
            if nulls_last:
                if _asc:
                    return col(_col).asc_nulls_last()
                else:
                    return col(_col).desc_nulls_last()
            else:
                if _asc:
                    return col(_col).asc_nulls_first()
                else:
                    return col(_col).desc_nulls_first()

        # If no partition
        if len(partition_spec.partition_by) == 0:
            if len(_presort.keys()) > 0:
                d = d.orderBy(
                    [_presort_to_col(_col, _presort[_col]) for _col in _presort.keys()]
                )
            d = d.limit(n)

        # If partition exists
        else:
            w = Window.partitionBy([col(x) for x in partition_spec.partition_by])

            if len(_presort.keys()) > 0:
                w = w.orderBy(
                    [_presort_to_col(_col, _presort[_col]) for _col in _presort.keys()]
                )
            else:
                # row_number() still needs an orderBy
                w = w.orderBy(lit(1))

            d = (
                d.select(col("*"), row_number().over(w).alias("__row_number__"))
                .filter(col("__row_number__") <= n)
                .drop("__row_number__")
            )

        return self._to_spark_df(d, df.schema)

    def load_df(
        self,
        path: Union[str, List[str]],
        format_hint: Any = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> DataFrame:
        return self._io.load_df(
            uri=path, format_hint=format_hint, columns=columns, **kwargs
        )

    def save_df(
        self,
        df: DataFrame,
        path: str,
        format_hint: Any = None,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        df = self._to_spark_df(df)
        self._io.save_df(
            df,
            uri=path,
            format_hint=format_hint,
            mode=mode,
            partition_spec=partition_spec,
            force_single=force_single,
            **kwargs,
        )

    def _broadcast(self, df: SparkDataFrame) -> SparkDataFrame:
        sdf = broadcast(df.native)
        return SparkDataFrame(sdf, df.schema)

    def _persist(self, df: SparkDataFrame, lazy: bool, level: Any) -> SparkDataFrame:
        if level is None:
            level = StorageLevel.MEMORY_AND_DISK
        if isinstance(level, str) and level in StorageLevel.__dict__:
            level = StorageLevel.__dict__[level]
        if isinstance(level, StorageLevel):
            df.native.persist()
            if not lazy:
                ct = df.count()
                self.log.info("Persist dataframe with %s, count %i", level, ct)
            return df
        raise ValueError(f"{level} is not supported persist type")  # pragma: no cover

    def _to_spark_df(  # noqa: C901
        self, df: Any, schema: Any = None, create_view: bool = False
    ) -> SparkDataFrame:
        def _to_df() -> SparkDataFrame:
            if isinstance(df, DataFrame):
                assert_or_throw(
                    schema is None,
                    ValueError("schema must be None when df is a DataFrame"),
                )
                if isinstance(df, SparkDataFrame):
                    return df
                if isinstance(df, (ArrayDataFrame, IterableDataFrame)):
                    adf = ArrowDataFrame(df.as_array(type_safe=False), df.schema)
                    sdf = to_spark_df(self.spark_session, adf, df.schema)
                    return SparkDataFrame(sdf, df.schema)
                if any(pa.types.is_struct(t) for t in df.schema.types):
                    sdf = to_spark_df(
                        self.spark_session, df.as_array(type_safe=True), df.schema
                    )
                else:
                    sdf = to_spark_df(self.spark_session, df, df.schema)
                return SparkDataFrame(sdf, df.schema)
            if is_spark_dataframe(df):
                return SparkDataFrame(df, None if schema is None else to_schema(schema))
            if isinstance(df, RDD):
                assert_arg_not_none(schema, "schema")
                sdf = to_spark_df(self.spark_session, df, schema)
                return SparkDataFrame(sdf, to_schema(schema))
            if isinstance(df, pd.DataFrame):
                if PD_UTILS.empty(df):
                    temp_schema = to_spark_schema(PD_UTILS.to_schema(df))
                    sdf = to_spark_df(self.spark_session, [], temp_schema)
                else:
                    sdf = to_spark_df(self.spark_session, df)
                return SparkDataFrame(sdf, schema)

            # use arrow dataframe here to handle nulls in int cols
            assert_or_throw(
                schema is not None, FugueDataFrameInitError("schema can't be None")
            )
            adf = ArrowDataFrame(df, to_schema(schema))
            map_pos = [i for i, t in enumerate(adf.schema.types) if pa.types.is_map(t)]
            if len(map_pos) == 0:
                sdf = to_spark_df(self.spark_session, adf.as_array(), adf.schema)
            else:

                def to_dict(rows: Iterable[List[Any]]) -> Iterable[List[Any]]:
                    for row in rows:
                        for p in map_pos:
                            row[p] = dict(row[p])
                        yield row

                sdf = to_spark_df(
                    self.spark_session, to_dict(adf.as_array_iterable()), adf.schema
                )
            return SparkDataFrame(sdf, adf.schema)

        res = _to_df()
        if res is not df and isinstance(df, DataFrame) and df.has_metadata:
            res.reset_metadata(df.metadata)

        if create_view:
            with self._lock:
                if res.alias not in self._registered_dfs:
                    res.native.createOrReplaceTempView(res.alias)
                    self._registered_dfs[res.alias] = res

        return res


class _Mapper(object):  # pragma: no cover
    # pytest can't identify the coverage, but this part is fully tested
    def __init__(
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]],
    ):
        super().__init__()
        self.schema = df.schema
        self.output_schema = Schema(output_schema)
        self.partition_spec = partition_spec
        self.map_func = map_func
        self.on_init = on_init
        self.cursor = self.partition_spec.get_cursor(self.schema, 0)

    def run(self, no: int, rows: Iterable[ps.Row]) -> Iterable[Any]:
        df = IterableDataFrame(to_type_safe_input(rows, self.schema), self.schema)
        if df.empty:  # pragma: no cover
            return
        if self.on_init is not None:
            self.on_init(no, df)
        if self.partition_spec.empty or self.partition_spec.algo == "coarse":
            partitions: Iterable[Tuple[int, int, EmptyAwareIterable]] = [
                (0, 0, df.native)
            ]
        else:
            partitioner = self.partition_spec.get_partitioner(self.schema)
            partitions = partitioner.partition(df.native)
        for pn, sn, sub in partitions:
            self.cursor.set(sub.peek(), pn, sn)
            sub_df = IterableDataFrame(sub, self.schema)
            res = self.map_func(self.cursor, sub_df)
            for r in res.as_array_iterable(type_safe=True):
                yield r


def _to_safe_spark_worker_pandas(df: pd.DataFrame):  # pragma: no cover
    return df if _PYSPARK_ARROW_FRIENDLY else df.astype(object)



from contextlib import contextmanager
from typing import Any, Callable, Iterator, List, Optional, Union

from triad import ParamDict, assert_or_throw

from fugue.column import ColumnExpr, SelectColumns, col, lit
from fugue.constants import _FUGUE_GLOBAL_CONF
from fugue.exceptions import FugueInvalidOperation

from ..collections.partition import PartitionSpec
from ..dataframe.dataframe import AnyDataFrame, DataFrame, as_fugue_df
from .execution_engine import (
    _FUGUE_GLOBAL_EXECUTION_ENGINE_CONTEXT,
    AnyExecutionEngine,
    ExecutionEngine,
)
from .factory import make_execution_engine, try_get_context_execution_engine
from .._utils.registry import fugue_plugin


@contextmanager
def engine_context(
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    infer_by: Optional[List[Any]] = None,
) -> Iterator[ExecutionEngine]:
    """Make an execution engine and set it as the context engine. This function
    is thread safe and async safe.

    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param infer_by: a list of objects to infer the engine, defaults to None

    .. note::

        For more details, please read
        :func:`~.fugue.execution.factory.make_execution_engine`

    .. admonition:: Examples

        .. code-block:: python

            import fugue.api as fa

            with fa.engine_context(spark_session):
                transform(df, func)  # will use spark in this transformation

    """
    e = make_execution_engine(engine, engine_conf, infer_by=infer_by)
    return e._as_context()


def set_global_engine(
    engine: AnyExecutionEngine, engine_conf: Any = None
) -> ExecutionEngine:
    """Make an execution engine and set it as the global execution engine

    :param engine: an engine like object, must not be None
    :param engine_conf: the configs for the engine, defaults to None

    .. caution::

        In general, it is not a good practice to set a global engine. You should
        consider :func:`~.engine_context` instead. The exception
        is when you iterate in a notebook and cross cells, this could simplify
        the code.

    .. note::

        For more details, please read
        :func:`~.fugue.execution.factory.make_execution_engine` and
        :meth:`~fugue.execution.execution_engine.ExecutionEngine.set_global`

    .. admonition:: Examples

        .. code-block:: python

            import fugue.api as fa

            fa.set_global_engine(spark_session)
            transform(df, func)  # will use spark in this transformation
            fa.clear_global_engine()  # remove the global setting
    """
    assert_or_throw(engine is not None, ValueError("engine must be specified"))
    return make_execution_engine(engine, engine_conf).set_global()


def clear_global_engine() -> None:
    """Remove the global exeuction engine (if set)"""
    _FUGUE_GLOBAL_EXECUTION_ENGINE_CONTEXT.set(None)


def get_context_engine() -> ExecutionEngine:
    """Get the execution engine in the current context. Regarding the order of the logic
    please read :func:`~.fugue.execution.factory.make_execution_engine`
    """
    engine = try_get_context_execution_engine()
    if engine is None:
        raise FugueInvalidOperation("No global/context engine is set")
    return engine


def get_current_conf() -> ParamDict:
    """Get the current configs either in the defined engine context or by
    the global configs (see :func:`~.fugue.constants.register_global_conf`)
    """
    engine = try_get_context_execution_engine()
    if engine is not None:
        return engine.conf
    return _FUGUE_GLOBAL_CONF


def get_current_parallelism() -> int:
    """Get the current parallelism of the current global/context engine.
    If there is no global/context engine, it creates a temporary engine
    using :func:`~.fugue.execution.factory.make_execution_engine` to get
    its parallelism

    :return: the size of the parallelism
    """
    return make_execution_engine().get_current_parallelism()


@fugue_plugin
def as_fugue_engine_df(
    engine: ExecutionEngine, df: AnyDataFrame, schema: Any = None
) -> DataFrame:
    """Convert a dataframe to a Fugue engine dependent DataFrame.
    This function is used internally by Fugue. It is not recommended
    to use

    :param engine: the ExecutionEngine to use, must not be None
    :param df: a dataframe like object
    :param schema: the schema of the dataframe, defaults to None

    :return: the engine dependent DataFrame
    """
    if schema is None:
        fdf = as_fugue_df(df)
    else:
        fdf = as_fugue_df(df, schema=schema)
    return engine.to_df(fdf)


def run_engine_function(
    func: Callable[[ExecutionEngine], Any],
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
    infer_by: Optional[List[Any]] = None,
) -> Any:
    """Run a lambda function based on the engine provided

    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False
    :param infer_by: a list of objects to infer the engine, defaults to None

    :return: None or a Fugue :class:`~.fugue.dataframe.dataframe.DataFrame` if
        ``as_fugue`` is True, otherwise if ``infer_by`` contains any
        Fugue DataFrame, then return the Fugue DataFrame, otherwise
        it returns the underlying dataframe using
        :meth:`~.fugue.dataframe.dataframe.DataFrame.native_as_df`

    .. note::

        This function is for deveopment use. Users should not need it.
    """
    with engine_context(engine, engine_conf=engine_conf, infer_by=infer_by) as e:
        res = func(e)

        if isinstance(res, DataFrame):
            res = e.convert_yield_dataframe(res, as_local=as_local)
            if as_fugue or any(isinstance(x, DataFrame) for x in (infer_by or [])):
                return res
            return res.native_as_df()
        return res


def repartition(
    df: AnyDataFrame,
    partition: PartitionSpec,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Partition the input dataframe using ``partition``.

    :param df: an input dataframe that can be recognized by Fugue
    :param partition: how you want to partition the dataframe
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the repartitioned dataframe

    .. caution::

        This function is experimental, and may be removed in the future.
    """
    return run_engine_function(
        lambda e: e.repartition(
            as_fugue_df(df), partition_spec=PartitionSpec(partition)
        ),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def broadcast(
    df: AnyDataFrame,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Broadcast the dataframe to all workers of a distributed computing backend

    :param df: an input dataframe that can be recognized by Fugue
    :param engine: an engine-like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the broadcasted dataframe
    """
    return run_engine_function(
        lambda e: e.broadcast(as_fugue_df(df)),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def persist(
    df: AnyDataFrame,
    lazy: bool = False,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
    **kwargs: Any,
) -> AnyDataFrame:
    """Force materializing and caching the dataframe

    :param df: an input dataframe that can be recognized by Fugue
    :param lazy: ``True``: first usage of the output will trigger persisting
        to happen; ``False`` (eager): persist is forced to happend immediately.
        Default to ``False``
    :param kwargs: parameter to pass to the underlying persist implementation
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the persisted dataframe
    """
    return run_engine_function(
        lambda e: e.persist(as_fugue_df(df), lazy=lazy, **kwargs),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def distinct(
    df: AnyDataFrame,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Equivalent to ``SELECT DISTINCT * FROM df``

    :param df: an input dataframe that can be recognized by Fugue
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the result with distinct rows
    """
    return run_engine_function(
        lambda e: e.distinct(as_fugue_df(df)),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def dropna(
    df: AnyDataFrame,
    how: str = "any",
    thresh: int = None,
    subset: List[str] = None,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Drop NA recods from dataframe

    :param df: an input dataframe that can be recognized by Fugue
    :param how: 'any' or 'all'. 'any' drops rows that contain any nulls.
        'all' drops rows that contain all nulls.
    :param thresh: int, drops rows that have less than thresh non-null values
    :param subset: list of columns to operate on
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: DataFrame with NA records dropped
    """
    return run_engine_function(
        lambda e: e.dropna(as_fugue_df(df), how=how, thresh=thresh, subset=subset),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def fillna(
    df: AnyDataFrame,
    value: Any,
    subset: List[str] = None,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """
    Fill ``NULL``, ``NAN``, ``NAT`` values in a dataframe

    :param df: an input dataframe that can be recognized by Fugue
    :param value: if scalar, fills all columns with same value.
        if dictionary, fills NA using the keys as column names and the
        values as the replacement values.
    :param subset: list of columns to operate on. ignored if value is
        a dictionary
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: DataFrame with NA records filled
    """
    return run_engine_function(
        lambda e: e.fillna(as_fugue_df(df), value=value, subset=subset),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def sample(
    df: AnyDataFrame,
    n: Optional[int] = None,
    frac: Optional[float] = None,
    replace: bool = False,
    seed: Optional[int] = None,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """
    Sample dataframe by number of rows or by fraction

    :param df: an input dataframe that can be recognized by Fugue
    :param n: number of rows to sample, one and only one of ``n`` and ``frac``
        must be set
    :param frac: fraction [0,1] to sample, one and only one of ``n`` and ``frac``
        must be set
    :param replace: whether replacement is allowed. With replacement,
        there may be duplicated rows in the result, defaults to False
    :param seed: seed for randomness, defaults to None
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the sampled dataframe
    """
    return run_engine_function(
        lambda e: e.sample(as_fugue_df(df), n=n, frac=frac, replace=replace, seed=seed),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def take(
    df: AnyDataFrame,
    n: int,
    presort: str,
    na_position: str = "last",
    partition: Any = None,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """
    Get the first n rows of a DataFrame per partition. If a presort is defined,
    use the presort before applying take. presort overrides partition_spec.presort.
    The Fugue implementation of the presort follows Pandas convention of specifying
    NULLs first or NULLs last. This is different from the Spark and SQL convention
    of NULLs as the smallest value.

    :param df: an input dataframe that can be recognized by Fugue
    :param n: number of rows to return
    :param presort: presort expression similar to partition presort
    :param na_position: position of null values during the presort.
        can accept ``first`` or ``last``
    :param partition: PartitionSpec to apply the take operation,
        defaults to None
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: n rows of DataFrame per partition
    """

    return run_engine_function(
        lambda e: e.take(
            as_fugue_df(df),
            n=n,
            presort=presort,
            na_position=na_position,
            partition_spec=None if partition is None else PartitionSpec(partition),
        ),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def load(
    path: Union[str, List[str]],
    format_hint: Any = None,
    columns: Any = None,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
    **kwargs: Any,
) -> AnyDataFrame:
    """Load dataframe from persistent storage

    :param path: the path to the dataframe
    :param format_hint: can accept ``parquet``, ``csv``, ``json``,
        defaults to None, meaning to infer
    :param columns: list of columns or a |SchemaLikeObject|, defaults to None
    :param kwargs: parameters to pass to the underlying framework
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False
    :return: an engine compatible dataframe

    For more details and examples, read |ZipComap|.
    """
    return run_engine_function(
        lambda e: e.load_df(
            path=path, format_hint=format_hint, columns=columns, **kwargs
        ),
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
    )


def save(
    df: AnyDataFrame,
    path: str,
    format_hint: Any = None,
    mode: str = "overwrite",
    partition: Any = None,
    force_single: bool = False,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    **kwargs: Any,
) -> None:
    """Save dataframe to a persistent storage

    :param df: an input dataframe that can be recognized by Fugue
    :param path: output path
    :param format_hint: can accept ``parquet``, ``csv``, ``json``,
        defaults to None, meaning to infer
    :param mode: can accept ``overwrite``, ``append``, ``error``,
        defaults to "overwrite"
    :param partition: how to partition the dataframe before saving,
        defaults to None
    :param force_single: force the output as a single file, defaults to False
    :param kwargs: parameters to pass to the underlying framework
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None

    For more details and examples, read |LoadSave|.
    """
    run_engine_function(
        lambda e: e.save_df(
            as_fugue_df(df),
            path=path,
            format_hint=format_hint,
            mode=mode,
            partition_spec=None if partition is None else PartitionSpec(partition),
            force_single=force_single,
            **kwargs,
        ),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
    )


def join(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    how: str,
    on: Optional[List[str]] = None,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Join two dataframes

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to join
    :param how: can accept ``semi``, ``left_semi``, ``anti``, ``left_anti``,
        ``inner``, ``left_outer``, ``right_outer``, ``full_outer``, ``cross``
    :param on: it can always be inferred, but if you provide, it will be
        validated against the inferred keys.
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the joined dataframe

    .. note::

        Please read :func:`~.fugue.dataframe.utils.get_join_schemas`
    """

    def _join(e: ExecutionEngine):
        edf1 = as_fugue_engine_df(e, df1)
        edf2 = as_fugue_engine_df(e, df2)
        res = e.join(edf1, edf2, how=how, on=on)
        for odf in dfs:
            res = e.join(res, as_fugue_engine_df(e, odf), how=how, on=on)
        return res

    return run_engine_function(
        _join,
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
        infer_by=[df1, df2, *dfs],
    )


def inner_join(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Inner join two dataframes.
    This is a wrapper of :func:`~.join` with ``how="inner"``

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to join
    :param how: can accept ``semi``, ``left_semi``, ``anti``, ``left_anti``,
        ``inner``, ``left_outer``, ``right_outer``, ``full_outer``, ``cross``
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the joined dataframe
    """
    return join(
        df1,
        df2,
        *dfs,
        how="inner",
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
    )


def semi_join(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Left semi-join two dataframes.
    This is a wrapper of :func:`~.join` with ``how="semi"``

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to join
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the joined dataframe
    """
    return join(
        df1,
        df2,
        *dfs,
        how="semi",
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
    )


def anti_join(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Left anti-join two dataframes.
    This is a wrapper of :func:`~.join` with ``how="anti"``

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to join
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the joined dataframe
    """
    return join(
        df1,
        df2,
        *dfs,
        how="anti",
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
    )


def left_outer_join(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Left outer join two dataframes.
    This is a wrapper of :func:`~.join` with ``how="left_outer"``

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to join
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the joined dataframe
    """
    return join(
        df1,
        df2,
        *dfs,
        how="left_outer",
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
    )


def right_outer_join(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Right outer join two dataframes.
    This is a wrapper of :func:`~.join` with ``how="right_outer"``

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to join
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the joined dataframe
    """
    return join(
        df1,
        df2,
        *dfs,
        how="right_outer",
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
    )


def full_outer_join(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Full outer join two dataframes.
    This is a wrapper of :func:`~.join` with ``how="full_outer"``

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to join
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the joined dataframe
    """
    return join(
        df1,
        df2,
        *dfs,
        how="full_outer",
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
    )


def cross_join(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Cross join two dataframes.
    This is a wrapper of :func:`~.join` with ``how="cross"``

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to join
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the joined dataframe
    """
    return join(
        df1,
        df2,
        *dfs,
        how="cross",
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
    )


def union(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    distinct: bool = True,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Join two dataframes

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to union
    :param distinct: ``true`` for ``UNION`` (== ``UNION DISTINCT``),
        ``false`` for ``UNION ALL``
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the unioned dataframe

    .. note::

        Currently, the schema of all dataframes must be identical, or
        an exception will be thrown.
    """

    def _union(e: ExecutionEngine):
        edf1 = as_fugue_engine_df(e, df1)
        edf2 = as_fugue_engine_df(e, df2)
        res = e.union(edf1, edf2, distinct=distinct)
        for odf in dfs:
            res = e.union(res, as_fugue_engine_df(e, odf), distinct=distinct)
        return res

    return run_engine_function(
        _union,
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
        infer_by=[df1, df2, *dfs],
    )


def subtract(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    distinct: bool = True,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """``df1 - df2``

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to subtract
    :param distinct: ``true`` for ``EXCEPT`` (== ``EXCEPT DISTINCT``),
        ``false`` for ``EXCEPT ALL``
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the unioned dataframe

    .. note::

        Currently, the schema of all datafrmes must be identical, or
        an exception will be thrown.
    """

    def _subtract(e: ExecutionEngine):
        edf1 = as_fugue_engine_df(e, df1)
        edf2 = as_fugue_engine_df(e, df2)
        res = e.subtract(edf1, edf2, distinct=distinct)
        for odf in dfs:
            res = e.subtract(res, as_fugue_engine_df(e, odf), distinct=distinct)
        return res

    return run_engine_function(
        _subtract,
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
        infer_by=[df1, df2, *dfs],
    )


def intersect(
    df1: AnyDataFrame,
    df2: AnyDataFrame,
    *dfs: AnyDataFrame,
    distinct: bool = True,  # pylint: disable-all
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Intersect ``df1`` and ``df2``

    :param df1: the first dataframe
    :param df2: the second dataframe
    :param dfs: more dataframes to intersect with
    :param distinct: ``true`` for ``INTERSECT`` (== ``INTERSECT DISTINCT``),
        ``false`` for ``INTERSECT ALL``
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the unioned dataframe

    .. note::

        Currently, the schema of ``df1`` and ``df2`` must be identical, or
        an exception will be thrown.
    """

    def _intersect(e: ExecutionEngine):
        edf1 = as_fugue_engine_df(e, df1)
        edf2 = as_fugue_engine_df(e, df2)
        res = e.intersect(edf1, edf2, distinct=distinct)
        for odf in dfs:
            res = e.intersect(res, as_fugue_engine_df(e, odf), distinct=distinct)
        return res

    return run_engine_function(
        _intersect,
        engine=engine,
        engine_conf=engine_conf,
        as_fugue=as_fugue,
        as_local=as_local,
        infer_by=[df1, df2, *dfs],
    )


def select(
    df: AnyDataFrame,
    *columns: Union[str, ColumnExpr],
    where: Optional[ColumnExpr] = None,
    having: Optional[ColumnExpr] = None,
    distinct: bool = False,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """The functional interface for SQL select statement

    :param df: the dataframe to be operated on
    :param columns: column expressions, for strings they will represent
        the column names
    :param where: ``WHERE`` condition expression, defaults to None
    :param having: ``having`` condition expression, defaults to None. It
        is used when ``cols`` contains aggregation columns, defaults to None
    :param distinct: whether to return distinct result, defaults to False
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the select result as a dataframe

    .. attention::

        This interface is experimental, it's subjected to change in new versions.

    .. seealso::

        Please find more expression examples in :mod:`fugue.column.sql` and
        :mod:`fugue.column.functions`

    .. admonition:: Examples

        .. code-block:: python

            from fugue.column import col, lit, functions as f
            import fugue.api as fa

            with fa.engine_context("duckdb"):
                # select existed and new columns
                fa.select(df, col("a"),col("b"),lit(1,"another"))
                fa.select(df, col("a"),(col("b")+lit(1)).alias("x"))

                # aggregation
                # SELECT COUNT(DISTINCT *) AS x FROM df
                fa.select(
                    df,
                    f.count_distinct(all_cols()).alias("x"))

                # SELECT a, MAX(b+1) AS x FROM df GROUP BY a
                fa.select(
                    df,
                    col("a"),f.max(col("b")+lit(1)).alias("x"))

                # SELECT a, MAX(b+1) AS x FROM df
                #   WHERE b<2 AND a>1
                #   GROUP BY a
                #   HAVING MAX(b+1)>0
                fa.select(
                    df,
                    col("a"),f.max(col("b")+lit(1)).alias("x"),
                    where=(col("b")<2) & (col("a")>1),
                    having=f.max(col("b")+lit(1))>0
                )
    """
    cols = SelectColumns(
        *[col(x) if isinstance(x, str) else x for x in columns],
        arg_distinct=distinct,
    )

    return run_engine_function(
        lambda e: e.select(as_fugue_df(df), cols=cols, where=where, having=having),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def filter(
    df: AnyDataFrame,
    condition: ColumnExpr,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Filter rows by the given condition

    :param df: the dataframe to be filtered
    :param condition: (boolean) column expression
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the filtered dataframe

    .. seealso::

        Please find more expression examples in :mod:`fugue.column.sql` and
        :mod:`fugue.column.functions`

    .. admonition:: Examples

        .. code-block:: python

            from fugue.column import col, functions as f
            import fugue.api as fa

            with fa.engine_context("duckdb"):
                fa.filter(df, (col("a")>1) & (col("b")=="x"))
                fa.filter(df, f.coalesce(col("a"),col("b"))>1)
    """
    return run_engine_function(
        lambda e: e.filter(as_fugue_df(df), condition=condition),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def assign(
    df: AnyDataFrame,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
    **columns: Any,
) -> AnyDataFrame:
    """Update existing columns with new values and add new columns

    :param df: the dataframe to set columns
    :param columns: column expressions
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the updated dataframe

    .. tip::

        This can be used to cast data types, alter column values or add new
        columns. But you can't use aggregation in columns.

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. seealso::

        Please find more expression examples in :mod:`fugue.column.sql` and
        :mod:`fugue.column.functions`

    .. admonition:: Examples

        .. code-block:: python

            from fugue.column import col, functions as f
            import fugue.api as fa

            # assume df has schema: a:int,b:str

            with fa.engine_context("duckdb"):
                # add constant column x
                fa.assign(df, x=1)

                # change column b to be a constant integer
                fa.assign(df, b=1)

                # add new x to be a+b
                fa.assign(df, x=col("a")+col("b"))

                # cast column a data type to double
                fa.assign(df, a=col("a").cast(float))
    """
    cols = [
        v.alias(k) if isinstance(v, ColumnExpr) else lit(v).alias(k)
        for k, v in columns.items()
    ]
    return run_engine_function(
        lambda e: e.assign(as_fugue_df(df), columns=cols),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )


def aggregate(
    df: AnyDataFrame,
    partition_by: Union[None, str, List[str]] = None,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
    **agg_kwcols: ColumnExpr,
) -> AnyDataFrame:
    """Aggregate on dataframe

    :param df: the dataframe to aggregate on
    :param partition_by: partition key(s), defaults to None
    :param agg_kwcols: aggregation expressions
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether to force return a local DataFrame, defaults to False

    :return: the aggregated result as a dataframe

    .. seealso::

        Please find more expression examples in :mod:`fugue.column.sql` and
        :mod:`fugue.column.functions`

    .. admonition:: Examples

        .. code-block:: python

            from fugue.column import col, functions as f
            import fugue.api as fa

            with fa.engine_context("duckdb"):
                # SELECT MAX(b) AS b FROM df
                fa.aggregate(df, b=f.max(col("b")))

                # SELECT a, MAX(b) AS x FROM df GROUP BY a
                fa.aggregate(df, "a", x=f.max(col("b")))
    """
    cols = [
        v.alias(k) if isinstance(v, ColumnExpr) else lit(v).alias(k)
        for k, v in agg_kwcols.items()
    ]
    return run_engine_function(
        lambda e: e.aggregate(
            as_fugue_df(df),
            partition_spec=None
            if partition_by is None
            else PartitionSpec(by=partition_by),
            agg_cols=cols,
        ),
        engine=engine,
        engine_conf=engine_conf,
        infer_by=[df],
        as_fugue=as_fugue,
        as_local=as_local,
    )



# flake8: noqa
from fugue.column.expressions import ColumnExpr, all_cols, col, function, lit, null
from fugue.column.functions import is_agg
from fugue.column.sql import SelectColumns, SQLExpressionGenerator



from typing import Any

from triad import assert_or_throw
from triad.utils.hash import to_uuid


class Yielded(object):
    """Yields from :class:`~fugue.workflow.workflow.FugueWorkflow`.
    Users shouldn't create this object directly.

    :param yid: unique id for determinism
    """

    def __init__(self, yid: str):
        self._yid = to_uuid(yid)

    def __uuid__(self) -> str:
        """uuid of the instance"""
        return self._yid

    @property
    def is_set(self) -> bool:  # pragma: no cover
        """Whether the value is set. It can be false if the parent workflow
        has not been executed.
        """
        raise NotImplementedError

    def __copy__(self) -> Any:  # pragma: no cover
        """``copy`` should have no effect"""
        return self

    def __deepcopy__(self, memo: Any) -> Any:  # pragma: no cover
        """``deepcopy`` should have no effect"""
        return self


class PhysicalYielded(Yielded):
    """Physical yielded object from :class:`~fugue.workflow.workflow.FugueWorkflow`.
    Users shouldn't create this object directly.

    :param yid: unique id for determinism
    :param storage_type: ``file`` or ``table``
    """

    def __init__(self, yid: str, storage_type: str):
        super().__init__(yid)
        self._name = ""
        assert_or_throw(
            storage_type in ["file", "table"],
            ValueError(f"{storage_type} not in (file, table) "),
        )
        self._storage_type = storage_type

    @property
    def is_set(self) -> bool:
        return self._name != ""

    def set_value(self, name: str) -> None:
        """Set the storage name after compute

        :param name: name reference of the storage
        """
        self._name = name

    @property
    def name(self) -> str:
        """The name reference of the yield"""
        assert_or_throw(self.is_set, "value is not set")
        return self._name

    @property
    def storage_type(self) -> str:
        """The storage type of this yield"""
        return self._storage_type



from typing import Any, Dict, Iterable, List, Optional, Tuple

import dask.dataframe as dd
import pandas as pd
import pyarrow as pa
from triad import assert_or_throw
from triad.collections.schema import Schema
from triad.utils.assertion import assert_arg_not_none
from triad.utils.pandas_like import PD_UTILS
from triad.utils.pyarrow import cast_pa_table

from fugue.dataframe import DataFrame, LocalBoundedDataFrame, PandasDataFrame
from fugue.dataframe.dataframe import _input_schema
from fugue.dataframe.pandas_dataframe import _pd_as_dicts
from fugue.exceptions import FugueDataFrameOperationError
from fugue.plugins import (
    as_array,
    as_array_iterable,
    as_dict_iterable,
    as_dicts,
    as_local_bounded,
    count,
    drop_columns,
    get_column_names,
    get_num_partitions,
    head,
    is_bounded,
    is_df,
    is_empty,
    is_local,
    rename,
    select_columns,
)

from ._constants import FUGUE_DASK_USE_ARROW
from ._utils import DASK_UTILS, collect, get_default_partitions


class DaskDataFrame(DataFrame):
    """DataFrame that wraps Dask DataFrame. Please also read
    |DataFrameTutorial| to understand this Fugue concept

    :param df: :class:`dask:dask.dataframe.DataFrame`,
      pandas DataFrame or list or iterable of arrays
    :param schema: |SchemaLikeObject| or :class:`spark:pyspark.sql.types.StructType`,
      defaults to None.
    :param num_partitions: initial number of partitions for the dask dataframe
      defaults to 0 to get the value from `fugue.dask.default.partitions`
    :param type_safe: whether to cast input data to ensure type safe, defaults to True

    .. note::

        For :class:`dask:dask.dataframe.DataFrame`, schema must be None
    """

    def __init__(  # noqa: C901
        self,
        df: Any = None,
        schema: Any = None,
        num_partitions: int = 0,
        type_safe=True,
    ):
        if num_partitions <= 0:
            num_partitions = get_default_partitions()
        if df is None:
            schema = _input_schema(schema).assert_not_empty()
            df = []
        if isinstance(df, DaskDataFrame):
            super().__init__(df.schema)
            self._native: dd.DataFrame = df._native
            return
        elif isinstance(df, (dd.DataFrame, dd.Series)):
            if isinstance(df, dd.Series):
                df = df.to_frame()
            pdf = df
            schema = None if schema is None else _input_schema(schema)
        elif isinstance(df, (pd.DataFrame, pd.Series)):
            if isinstance(df, pd.Series):
                df = df.to_frame()
            schema = None if schema is None else _input_schema(schema)
            tdf = PandasDataFrame(df, schema=schema)
            pdf = dd.from_pandas(tdf.native, npartitions=num_partitions, sort=False)
            schema = tdf.schema
            type_safe = False
        elif isinstance(df, Iterable):
            schema = _input_schema(schema).assert_not_empty()
            tdf = PandasDataFrame(df, schema=schema)
            pdf = dd.from_pandas(tdf.native, npartitions=num_partitions, sort=False)
            schema = tdf.schema
            type_safe = False
        else:
            raise ValueError(f"{df} is incompatible with DaskDataFrame")
        pdf, schema = self._apply_schema(pdf, schema, type_safe)
        super().__init__(schema)
        self._native = pdf

    @property
    def native(self) -> dd.DataFrame:
        """The wrapped Dask DataFrame"""
        return self._native

    def native_as_df(self) -> dd.DataFrame:
        return self._native

    @property
    def is_local(self) -> bool:
        return False

    def as_local_bounded(self) -> LocalBoundedDataFrame:
        res = PandasDataFrame(self.as_pandas(), self.schema)
        if self.has_metadata:
            res.reset_metadata(self.metadata)
        return res

    @property
    def is_bounded(self) -> bool:
        return True

    @property
    def empty(self) -> bool:
        return DASK_UTILS.empty(self.native)

    @property
    def num_partitions(self) -> int:
        return _dd_get_num_partitions(self.native)

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        cols = (self.schema - cols).names
        return self._select_cols(cols)

    def _select_cols(self, cols: List[Any]) -> DataFrame:
        schema = self.schema.extract(cols)
        return DaskDataFrame(self.native[schema.names], schema, type_safe=False)

    def peek_array(self) -> List[Any]:
        self.assert_not_empty()
        return self.as_pandas().iloc[0].values.tolist()

    def persist(self, **kwargs: Any) -> "DaskDataFrame":
        self._native = self.native.persist(**kwargs)
        self._native.count().compute()
        return self

    def count(self) -> int:
        return self.native.shape[0].compute()

    def as_pandas(self) -> pd.DataFrame:
        pdf = self.native.compute().reset_index(drop=True)
        return PD_UTILS.cast_df(
            pdf, self.schema.pa_schema, use_extension_types=True, use_arrow_dtype=False
        )

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        schema = self.schema.pa_schema
        return pa.concat_tables(
            collect(
                self.native,
                lambda df: cast_pa_table(
                    pa.Table.from_pandas(df.reset_index(drop=True), schema=schema),
                    schema=schema,
                ),
            )
        )

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        try:
            schema = self.schema.rename(columns)
        except Exception as e:
            raise FugueDataFrameOperationError from e
        df = self.native.rename(columns=columns)
        return DaskDataFrame(df, schema, type_safe=False)

    def alter_columns(self, columns: Any) -> DataFrame:
        new_schema = self.schema.alter(columns)
        if new_schema == self.schema:
            return self
        return DaskDataFrame(self.native, new_schema)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        chunks = _to_array_chunks(self.native, columns, type_safe, schema=self.schema)
        res: List[List[Any]] = []
        for x in chunks:
            res += x
        return res

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        chunks = _to_array_chunks(self.native, columns, type_safe, schema=self.schema)
        for x in chunks:
            yield from x

    def as_dicts(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Dict[str, Any]]:
        return _dd_as_dicts(self.native, columns)

    def as_dict_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Dict[str, Any]]:
        yield from _dd_as_dict_iterable(self.native, columns)

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        ddf = self.native if columns is None else self.native[columns]
        schema = self.schema if columns is None else self.schema.extract(columns)
        return PandasDataFrame(ddf.head(n, compute=True, npartitions=-1), schema=schema)

    def _apply_schema(
        self, pdf: dd.DataFrame, schema: Optional[Schema], type_safe: bool = True
    ) -> Tuple[dd.DataFrame, Schema]:
        if not type_safe:
            assert_arg_not_none(pdf, "pdf")
            assert_arg_not_none(schema, "schema")
            return pdf, schema
        DASK_UTILS.ensure_compatible(pdf)
        # when pdf contains bytes, or any object types, and schema contains str
        # there is no way to get the real schema of the pdf, (pschema will contain
        # strs instead of the real types) so we have to force cast it to the schema
        if schema is None:
            pschema = Schema(DASK_UTILS.to_schema(pdf))
            return pdf, pschema.assert_not_empty()
        pdf = pdf[schema.assert_not_empty().names]
        return (
            DASK_UTILS.cast_df(
                pdf,
                schema.pa_schema,
                use_extension_types=True,
                use_arrow_dtype=FUGUE_DASK_USE_ARROW,
            ),
            schema,
        )


@is_df.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_is_df(df: dd.DataFrame) -> bool:
    return True


@get_num_partitions.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_get_num_partitions(df: dd.DataFrame) -> int:
    return df.npartitions


@count.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_count(df: dd.DataFrame) -> int:
    return df.shape[0].compute()


@is_bounded.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_is_bounded(df: dd.DataFrame) -> bool:
    return True


@is_empty.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_is_empty(df: dd.DataFrame) -> bool:
    return DASK_UTILS.empty(df)


@is_local.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_is_local(df: dd.DataFrame) -> bool:
    return False


@as_local_bounded.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_as_local(df: dd.DataFrame) -> bool:
    return df.compute()


@get_column_names.candidate(lambda df: isinstance(df, dd.DataFrame))
def _get_dask_dataframe_columns(df: dd.DataFrame) -> List[Any]:
    return list(df.columns)


@rename.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _rename_dask_dataframe(df: dd.DataFrame, columns: Dict[str, Any]) -> dd.DataFrame:
    if len(columns) == 0:
        return df
    _assert_no_missing(df, columns.keys())
    return df.rename(columns=columns)


@drop_columns.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _drop_dd_columns(
    df: dd.DataFrame, columns: List[str], as_fugue: bool = False
) -> Any:
    cols = [x for x in df.columns if x not in columns]
    if len(cols) == 0:
        raise FugueDataFrameOperationError("cannot drop all columns")
    if len(cols) + len(columns) != len(df.columns):
        _assert_no_missing(df, columns)
    return _adjust_df(df[cols], as_fugue=as_fugue)


@select_columns.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _select_dd_columns(
    df: dd.DataFrame, columns: List[Any], as_fugue: bool = False
) -> Any:
    if len(columns) == 0:
        raise FugueDataFrameOperationError("must select at least one column")
    _assert_no_missing(df, columns)
    return _adjust_df(df[columns], as_fugue=as_fugue)


@head.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _dd_head(
    df: dd.DataFrame,
    n: int,
    columns: Optional[List[str]] = None,
    as_fugue: bool = False,
) -> pd.DataFrame:
    if columns is not None:
        df = df[columns]
    res = df.head(n, compute=True, npartitions=-1)
    return PandasDataFrame(res) if as_fugue else res


@as_array.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _dd_as_array(
    df: dd.DataFrame, columns: Optional[List[str]] = None, type_safe: bool = False
) -> List[Any]:
    chunks = _to_array_chunks(df, columns, type_safe)
    res: List[List[Any]] = []
    for x in chunks:
        res += x
    return res


@as_array_iterable.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _dd_as_array_iterable(
    df: dd.DataFrame, columns: Optional[List[str]] = None, type_safe: bool = False
) -> Iterable[Any]:
    chunks = _to_array_chunks(df, columns, type_safe)
    for x in chunks:
        yield from x


@as_dicts.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _dd_as_dicts(
    df: dd.DataFrame, columns: Optional[List[str]] = None, type_safe: bool = False
) -> List[Dict[str, Any]]:
    assert_or_throw(columns is None or len(columns) > 0, ValueError("empty columns"))
    _df = df if columns is None or len(columns) == 0 else df[columns]
    res: List[Dict[str, Any]] = []
    for x in collect(_df, lambda df: _pd_as_dicts(df, columns)):
        res += x
    return res


@as_dict_iterable.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _dd_as_dict_iterable(
    df: dd.DataFrame, columns: Optional[List[str]] = None, type_safe: bool = False
) -> Iterable[Dict[str, Any]]:
    assert_or_throw(columns is None or len(columns) > 0, ValueError("empty columns"))
    _df = df if columns is None or len(columns) == 0 else df[columns]
    for x in collect(_df, lambda df: _pd_as_dicts(df, columns)):
        yield from x


def _assert_no_missing(df: dd.DataFrame, columns: Iterable[Any]) -> None:
    missing = set(columns) - set(df.columns)
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")


def _adjust_df(res: dd.DataFrame, as_fugue: bool):
    return res if not as_fugue else DaskDataFrame(res)


def _to_array_chunks(
    df: dd.DataFrame,
    columns: Optional[List[str]] = None,
    type_safe: bool = False,
    schema: Optional[Schema] = None,
) -> Tuple[List[Any]]:
    assert_or_throw(columns is None or len(columns) > 0, ValueError("empty columns"))
    _df = df if columns is None or len(columns) == 0 else df[columns]

    def _to_list(pdf: pd.DataFrame) -> List[Any]:
        return list(
            PD_UTILS.as_array_iterable(
                pdf,
                schema=None if schema is None else schema.pa_schema,
                columns=columns,
                type_safe=type_safe,
            )
        )

    return collect(_df, _to_list)



from typing import List, no_type_check

from triad import ParamDict, Schema, SerializableRLock, assert_or_throw
from triad.utils.convert import to_type

from fugue.collections.partition import PartitionCursor
from fugue.dataframe import DataFrame, DataFrames, LocalDataFrame
from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.dataframe.utils import _df_eq
from fugue.exceptions import FugueWorkflowError
from fugue.execution.execution_engine import (
    _FUGUE_SERIALIZED_BLOB_SCHEMA,
    _generate_comap_empty_dfs,
)
from fugue.rpc import EmptyRPCHandler, to_rpc_handler

from ..outputter import Outputter
from ..transformer.convert import _to_output_transformer
from ..transformer.transformer import CoTransformer, Transformer


class Show(Outputter):
    LOCK = SerializableRLock()

    def process(self, dfs: DataFrames) -> None:
        title = self.params.get_or_none("title", object)
        title = str(title) if title is not None else None
        n = self.params.get("n", 10)
        with_count = self.params.get("with_count", False)
        with Show.LOCK:
            m = 0
            for df in dfs.values():
                df.show(n=n, with_count=with_count, title=title if m == 0 else None)
                m += 1


class AssertEqual(Outputter):
    def process(self, dfs: DataFrames) -> None:
        assert_or_throw(len(dfs) > 1, FugueWorkflowError("can't accept single input"))
        expected = dfs[0].as_local_bounded()
        for i in range(1, len(dfs)):
            _df_eq(expected, dfs[i], throw=True, **self.params)


class AssertNotEqual(Outputter):
    def process(self, dfs: DataFrames) -> None:
        assert_or_throw(len(dfs) > 1, FugueWorkflowError("can't accept single input"))
        expected = dfs[0]
        for i in range(1, len(dfs)):
            assert not _df_eq(expected, dfs[i], throw=False, **self.params)


class Save(Outputter):
    def process(self, dfs: DataFrames) -> None:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        kwargs = self.params.get("params", dict())
        path = self.params.get_or_throw("path", str)
        format_hint = self.params.get("fmt", "")
        mode = self.params.get("mode", "overwrite")
        partition_spec = self.partition_spec
        force_single = self.params.get("single", False)

        self.execution_engine.save_df(
            df=dfs[0],
            path=path,
            format_hint=format_hint,
            mode=mode,
            partition_spec=partition_spec,
            force_single=force_single,
            **kwargs
        )


class RunOutputTransformer(Outputter):
    @no_type_check
    def process(self, dfs: DataFrames) -> None:
        df = dfs[0]
        tf = _to_output_transformer(
            self.params.get_or_none("transformer", object),
        )
        tf._workflow_conf = self.execution_engine.conf
        tf._params = self.params.get("params", ParamDict())  # type: ignore
        tf._partition_spec = self.partition_spec  # type: ignore
        rpc_handler = to_rpc_handler(self.params.get_or_throw("rpc_handler", object))
        if not isinstance(rpc_handler, EmptyRPCHandler):
            tf._rpc_client = self.rpc_server.make_client(rpc_handler)
        ie = self.params.get("ignore_errors", [])
        self._ignore_errors = [to_type(x, Exception) for x in ie]
        tf.validate_on_runtime(df)
        if isinstance(tf, Transformer):
            self.transform(df, tf)
        else:
            self.cotransform(df, tf)

    def transform(self, df: DataFrame, tf: Transformer) -> None:
        tf._key_schema = self.partition_spec.get_key_schema(df.schema)  # type: ignore
        tf._output_schema = Schema(tf.get_output_schema(df))  # type: ignore
        tr = _TransformerRunner(df, tf, self._ignore_errors)  # type: ignore
        df = self.execution_engine.map_engine.map_dataframe(
            df=df,
            map_func=tr.run,
            output_schema=tf.output_schema,  # type: ignore
            partition_spec=tf.partition_spec,
            on_init=tr.on_init,
            map_func_format_hint=tf.get_format_hint(),
        )
        self.execution_engine.persist(df, lazy=False)

    @no_type_check
    def cotransform(self, df: DataFrame, tf: CoTransformer) -> None:
        assert_or_throw(
            df.metadata.get("serialized", False), "must use serialized dataframe"
        )
        tf._key_schema = df.schema - _FUGUE_SERIALIZED_BLOB_SCHEMA
        # TODO: currently, get_output_schema only gets empty dataframes
        empty_dfs = _generate_comap_empty_dfs(
            df.metadata["schemas"], df.metadata.get("serialized_has_name", False)
        )
        tf._output_schema = Schema(tf.get_output_schema(empty_dfs))
        tr = _CoTransformerRunner(df, tf, self._ignore_errors)
        df = self.execution_engine.comap(
            df=df,
            map_func=tr.run,
            output_schema=tf.output_schema,
            partition_spec=tf.partition_spec,
            on_init=tr.on_init,
        )
        self.execution_engine.persist(df, lazy=False)


class _TransformerRunner(object):
    def __init__(
        self, df: DataFrame, transformer: Transformer, ignore_errors: List[type]
    ):
        self.schema = df.schema
        self.transformer = transformer
        self.ignore_errors = tuple(ignore_errors)

    def run(self, cursor: PartitionCursor, df: LocalDataFrame) -> LocalDataFrame:
        self.transformer._cursor = cursor  # type: ignore
        try:
            self.transformer.transform(df).as_local_bounded()
            return ArrayDataFrame([], self.transformer.output_schema)
        except self.ignore_errors:  # type: ignore
            return ArrayDataFrame([], self.transformer.output_schema)

    def on_init(self, partition_no: int, df: DataFrame) -> None:
        s = self.transformer.partition_spec
        self.transformer._cursor = s.get_cursor(  # type: ignore
            self.schema, partition_no
        )
        self.transformer.on_init(df)


class _CoTransformerRunner(object):
    def __init__(
        self, df: DataFrame, transformer: CoTransformer, ignore_errors: List[type]
    ):
        self.schema = df.schema
        self.transformer = transformer
        self.ignore_errors = tuple(ignore_errors)

    def run(self, cursor: PartitionCursor, dfs: DataFrames) -> LocalDataFrame:
        self.transformer._cursor = cursor  # type: ignore
        try:
            self.transformer.transform(dfs).as_local_bounded()
            return ArrayDataFrame([], self.transformer.output_schema)
        except self.ignore_errors:  # type: ignore
            return ArrayDataFrame([], self.transformer.output_schema)

    def on_init(self, partition_no: int, dfs: DataFrames) -> None:
        s = self.transformer.partition_spec
        self.transformer._cursor = s.get_cursor(  # type: ignore
            self.schema, partition_no
        )
        self.transformer.on_init(dfs)



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd
from dask import dataframe as dd
from fsspec import AbstractFileSystem
from triad.collections.dict import ParamDict
from triad.collections.schema import Schema
from triad.utils.assertion import assert_or_throw
from triad.utils.io import join, makedirs, url_to_fs

from fugue._utils.io import FileParser, _get_single_files
from fugue_dask.dataframe import DaskDataFrame

from ._utils import DASK_UTILS


def load_df(
    uri: Union[str, List[str]],
    format_hint: Optional[str] = None,
    columns: Any = None,
    fs: Optional[AbstractFileSystem] = None,
    **kwargs: Any,
) -> DaskDataFrame:
    if isinstance(uri, str):
        fp = [FileParser(uri, format_hint)]
    else:
        fp = [FileParser(u, format_hint) for u in uri]
    dfs: List[dd.DataFrame] = []
    schema: Any = None
    for f in _get_single_files(fp, fs):
        df, schema = _FORMAT_LOAD[f.file_format](f, columns, **kwargs)
        dfs.append(df)
    return DaskDataFrame(dd.concat(dfs), schema)


def save_df(
    df: DaskDataFrame,
    uri: str,
    format_hint: Optional[str] = None,
    mode: str = "overwrite",
    fs: Optional[AbstractFileSystem] = None,
    **kwargs: Any,
) -> None:
    assert_or_throw(
        mode in ["overwrite", "error"],
        lambda: NotImplementedError(f"{mode} is not supported"),
    )
    p = FileParser(uri, format_hint).assert_no_glob()
    if fs is None:
        fs, _ = url_to_fs(uri)
    if fs.exists(uri):
        assert_or_throw(mode == "overwrite", FileExistsError(uri))
        try:
            fs.rm(uri, recursive=True)
        except Exception:  # pragma: no cover
            pass
    _FORMAT_SAVE[p.file_format](df, p, **kwargs)


def _save_parquet(df: DaskDataFrame, p: FileParser, **kwargs: Any) -> None:
    params = {
        "schema": df.schema.pa_schema,
        "write_index": False,
        **kwargs,
    }
    DASK_UTILS.to_parquet_friendly(df.native).to_parquet(p.path, **params)


def _load_parquet(
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[dd.DataFrame, Any]:
    params = dict(kwargs)
    dtype_backend: Any = params.pop("dtype_backend", None)
    params["engine"] = "pyarrow"
    params["arrow_to_pandas"] = {"ignore_metadata": True}
    if pd.__version__ >= "1.5":
        dtype_backend = "pyarrow"
    if columns is None:
        pdf = dd.read_parquet(p.path, dtype_backend=dtype_backend, **params)
        schema = Schema(pdf.head(1))
        return pdf, schema
    if isinstance(columns, list):  # column names
        pdf = dd.read_parquet(
            p.path, columns=columns, dtype_backend=dtype_backend, **params
        )
        schema = Schema(pdf.head(1))
        return pdf, schema
    schema = Schema(columns)
    pdf = dd.read_parquet(
        p.path, columns=schema.names, dtype_backend=dtype_backend, **params
    )
    return pdf, schema


def _save_csv(df: DaskDataFrame, p: FileParser, **kwargs: Any) -> None:
    makedirs(p.path, exist_ok=True)
    df.native.to_csv(
        p.join("*.csv").path, **{"index": False, "header": False, **kwargs}
    )


def _safe_load_csv(path: str, **kwargs: Any) -> dd.DataFrame:
    try:
        return dd.read_csv(path, **kwargs)
    except (IsADirectoryError, PermissionError):
        return dd.read_csv(join(path, "*.csv"), **kwargs)


def _load_csv(  # noqa: C901
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[dd.DataFrame, Any]:
    kw = ParamDict(kwargs)
    infer_schema = kw.get("infer_schema", False)
    if infer_schema and columns is not None and not isinstance(columns, list):
        raise ValueError("can't set columns as a schema when infer schema is true")
    if not infer_schema:
        kw["dtype"] = object
    if "infer_schema" in kw:
        del kw["infer_schema"]
    header: Any = False
    if "header" in kw:
        header = kw["header"]
        del kw["header"]
    if str(header) in ["True", "0"]:
        pdf = _safe_load_csv(p.path, **{"header": 0, **kw})
        if columns is None:
            return pdf, None
        if isinstance(columns, list):  # column names
            return pdf[columns], None
        schema = Schema(columns)
        return pdf[schema.names], schema
    if header is None or str(header) == "False":
        if columns is None:
            raise ValueError("columns must be set if without header")
        if isinstance(columns, list):  # column names
            pdf = _safe_load_csv(p.path, **{"header": None, "names": columns, **kw})
            return pdf, None
        schema = Schema(columns)
        pdf = _safe_load_csv(p.path, **{"header": None, "names": schema.names, **kw})
        return pdf, schema
    else:
        raise NotImplementedError(f"{header} is not supported")


def _save_json(df: DaskDataFrame, p: FileParser, **kwargs: Any) -> None:
    makedirs(p.path, exist_ok=True)
    df.native.to_json(p.join("*.json").path, **kwargs)


def _safe_load_json(path: str, **kwargs: Any) -> dd.DataFrame:
    try:
        return dd.read_json(path, **kwargs)
    except (IsADirectoryError, PermissionError):
        x = dd.read_json(join(path, "*.json"), **kwargs)
        return x


def _load_json(
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[dd.DataFrame, Any]:
    pdf = _safe_load_json(p.path, **kwargs).reset_index(drop=True)
    if columns is None:
        return pdf, None
    if isinstance(columns, list):  # column names
        return pdf[columns], None
    schema = Schema(columns)
    return pdf[schema.names], schema


_FORMAT_LOAD: Dict[str, Callable[..., Tuple[dd.DataFrame, Any]]] = {
    "csv": _load_csv,
    "parquet": _load_parquet,
    "json": _load_json,
}

_FORMAT_SAVE: Dict[str, Callable] = {
    "csv": _save_csv,
    "parquet": _save_parquet,
    "json": _save_json,
}



from fugue import ExecutionEngine

from .._constants import FUGUE_RAY_CONF_SHUFFLE_PARTITIONS, FUGUE_RAY_DEFAULT_PARTITIONS
from fugue.constants import FUGUE_CONF_DEFAULT_PARTITIONS


def get_default_partitions(engine: ExecutionEngine) -> int:
    n = engine.conf.get(
        FUGUE_RAY_DEFAULT_PARTITIONS, engine.conf.get(FUGUE_CONF_DEFAULT_PARTITIONS, -1)
    )
    return n if n >= 0 else engine.get_current_parallelism() * 2


def get_default_shuffle_partitions(engine: ExecutionEngine) -> int:
    n = engine.conf.get(FUGUE_RAY_CONF_SHUFFLE_PARTITIONS, -1)
    return n if n >= 0 else get_default_partitions(engine)



import uuid

import pytest
from triad.utils.io import makedirs, rm


@pytest.fixture
def tmp_mem_dir():
    uuid_str = str(uuid.uuid4())[:5]
    path = "memory://test_" + uuid_str
    makedirs(path)
    try:
        yield path
    finally:
        try:
            rm(path, recursive=True)
        except Exception:  # pragma: no cover
            pass



# flake8: noqa
# pylint: disable-all

try:  # pragma: no cover
    from ibis.expr.types import Table as IbisTable
except Exception:  # pragma: no cover
    from ibis.expr.types import TableExpr as IbisTable

from ibis import Schema as IbisSchema



from typing import Any, Dict, Tuple, Optional

from triad.utils.convert import get_caller_global_local_vars

from fugue.dataframe import AnyDataFrame
from fugue.exceptions import FugueSQLError
from fugue.execution import AnyExecutionEngine
from fugue.execution.api import get_current_conf

from ..constants import (
    FUGUE_CONF_SQL_IGNORE_CASE,
    FUGUE_CONF_SQL_DIALECT,
    FUGUE_SQL_DEFAULT_DIALECT,
)
from .workflow import FugueSQLWorkflow


def fugue_sql(
    query: str,
    *args: Any,
    fsql_ignore_case: Optional[bool] = None,
    fsql_dialect: Optional[str] = None,
    engine: AnyExecutionEngine = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
    **kwargs: Any,
) -> AnyDataFrame:
    """Simplified Fugue SQL interface. This function can still take multiple dataframe
    inputs but will always return the last generated dataframe in the SQL workflow. And
    ``YIELD`` should NOT be used with this function. If you want to use Fugue SQL to
    represent the full workflow, or want to see more Fugue SQL examples,
    please read :func:`~.fugue_sql_flow`.

    :param query: the Fugue SQL string (can be a jinja template)
    :param args: variables related to the SQL string
    :param fsql_ignore_case: whether to ignore case when parsing the SQL string,
        defaults to None (it depends on the engine/global config).
    :param fsql_dialect: the dialect of this fsql,
        defaults to None (it depends on the engine/global config).
    :param kwargs: variables related to the SQL string
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether return a local dataframe, defaults to False

    :return: the result dataframe

    .. note::

        This function is different from :func:`~fugue.api.raw_sql` which directly
        sends the query to the execution engine to run. This function parses the query
        based on Fugue SQL syntax, creates a
        :class:`~fugue.sql.workflow.FugueSQLWorkflow` which
        could contain multiple raw SQLs plus other operations, and runs and returns
        the last dataframe generated in the workflow.

        This function allows you to parameterize the SQL in a more elegant way. The
        data tables referred in the query can either be automatically extracted from the
        local variables or be specified in the arguments.

    .. caution::

        Currently, we have not unified the dialects of different SQL backends. So there
        can be some slight syntax differences when you switch between backends.
        In addition, we have not unified the UDFs cross different backends, so you
        should be careful to use uncommon UDFs belonging to a certain backend.

        That being said, if you keep your SQL part general and leverage Fugue extensions
        (transformer, creator, processor, outputter, etc.) appropriately, it should be
        easy to write backend agnostic Fugue SQL.

        We are working on unifying the dialects of different SQLs, it should be
        available in the future releases. Regarding unifying UDFs, the effort is still
        unclear.

    .. code-block:: python

        import pandas as pd
        import fugue.api as fa

        def tr(df:pd.DataFrame) -> pd.DataFrame:
            return df.assign(c=2)

        input = pd.DataFrame([[0,1],[3.4]], columns=["a","b"])

        with fa.engine_context("duckdb"):
            res = fa.fugue_sql('''
            SELECT * FROM input WHERE a<{{x}}
            TRANSFORM USING tr SCHEMA *,c:int
            ''', x=2)
            assert fa.as_array(res) == [[0,1,2]]
    """

    dag = _build_dag(
        query,
        fsql_ignore_case=fsql_ignore_case,
        fsql_dialect=fsql_dialect,
        args=args,
        kwargs=kwargs,
    )
    if dag.last_df is not None:
        dag.last_df.yield_dataframe_as("result", as_local=as_local)
    else:  # pragma: no cover
        # impossible case
        raise FugueSQLError(f"no dataframe to output from\n{query}")
    res = dag.run(engine, engine_conf)
    return res["result"] if as_fugue else res["result"].native_as_df()


def fugue_sql_flow(
    query: str,
    *args: Any,
    fsql_ignore_case: Optional[bool] = None,
    fsql_dialect: Optional[str] = None,
    **kwargs: Any,
) -> FugueSQLWorkflow:
    """Fugue SQL full functional interface. This function allows full workflow
    definition using Fugue SQL, and it allows multiple outputs using ``YIELD``.

    :param query: the Fugue SQL string (can be a jinja template)
    :param args: variables related to the SQL string
    :param fsql_ignore_case: whether to ignore case when parsing the SQL string,
        defaults to None (it depends on the engine/global config).
    :param fsql_dialect: the dialect of this fsql,
        defaults to None (it depends on the engine/global config).
    :param kwargs: variables related to the SQL string
    :return: the translated Fugue workflow

    .. note::

        This function is different from :func:`~fugue.api.raw_sql` which directly
        sends the query to the execution engine to run. This function parses the query
        based on Fugue SQL syntax, creates a
        :class:`~fugue.sql.workflow.FugueSQLWorkflow` which
        could contain multiple raw SQLs plus other operations, and runs and returns
        the last dataframe generated in the workflow.

        This function allows you to parameterize the SQL in a more elegant way. The
        data tables referred in the query can either be automatically extracted from the
        local variables or be specified in the arguments.

    .. caution::

        Currently, we have not unified the dialects of different SQL backends. So there
        can be some slight syntax differences when you switch between backends.
        In addition, we have not unified the UDFs cross different backends, so you
        should be careful to use uncommon UDFs belonging to a certain backend.

        That being said, if you keep your SQL part general and leverage Fugue extensions
        (transformer, creator, processor, outputter, etc.) appropriately, it should be
        easy to write backend agnostic Fugue SQL.

        We are working on unifying the dialects of different SQLs, it should be
        available in the future releases. Regarding unifying UDFs, the effort is still
        unclear.

    .. code-block:: python

        import fugue.api.fugue_sql_flow as fsql
        import fugue.api as fa

        # Basic case
        fsql('''
        CREATE [[0]] SCHEMA a:int
        PRINT
        ''').run()

        # With external data sources
        df = pd.DataFrame([[0],[1]], columns=["a"])
        fsql('''
        SELECT * FROM df WHERE a=0
        PRINT
        ''').run()

        # With external variables
        df = pd.DataFrame([[0],[1]], columns=["a"])
        t = 1
        fsql('''
        SELECT * FROM df WHERE a={{t}}
        PRINT
        ''').run()

        # The following is the explicit way to specify variables and datafrems
        # (recommended)
        df = pd.DataFrame([[0],[1]], columns=["a"])
        t = 1
        fsql('''
        SELECT * FROM df WHERE a={{t}}
        PRINT
        ''', df=df, t=t).run()

        # Using extensions
        def dummy(df:pd.DataFrame) -> pd.DataFrame:
            return df

        fsql('''
        CREATE [[0]] SCHEMA a:int
        TRANSFORM USING dummy SCHEMA *
        PRINT
        ''').run()

        # It's recommended to provide full path of the extension inside
        # Fugue SQL, so the SQL definition and exeuction can be more
        # independent from the extension definition.

        # Run with different execution engines
        sql = '''
        CREATE [[0]] SCHEMA a:int
        TRANSFORM USING dummy SCHEMA *
        PRINT
        '''

        fsql(sql).run(spark_session)
        fsql(sql).run("dask")

        with fa.engine_context("duckdb"):
            fsql(sql).run()

        # Passing dataframes between fsql calls
        result = fsql('''
        CREATE [[0]] SCHEMA a:int
        YIELD DATAFRAME AS x

        CREATE [[1]] SCHEMA a:int
        YIELD DATAFRAME AS y
        ''').run(DaskExecutionEngine)

        fsql('''
        SELECT * FROM x
        UNION
        SELECT * FROM y
        UNION
        SELECT * FROM z

        PRINT
        ''', result, z=pd.DataFrame([[2]], columns=["z"])).run()

        # Get framework native dataframes
        result["x"].native  # Dask dataframe
        result["y"].native  # Dask dataframe
        result["x"].as_pandas()  # Pandas dataframe

        # Use lower case fugue sql
        df = pd.DataFrame([[0],[1]], columns=["a"])
        t = 1
        fsql('''
        select * from df where a={{t}}
        print
        ''', df=df, t=t, fsql_ignore_case=True).run()
    """
    dag = _build_dag(
        query,
        fsql_ignore_case=fsql_ignore_case,
        fsql_dialect=fsql_dialect,
        args=args,
        kwargs=kwargs,
    )
    return dag


def _build_dag(
    query: str,
    fsql_ignore_case: Optional[bool],
    fsql_dialect: Optional[str],
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
    level: int = -2,
) -> FugueSQLWorkflow:
    global_vars, local_vars = get_caller_global_local_vars(start=level, end=level)
    if fsql_ignore_case is None:
        fsql_ignore_case = get_current_conf().get(FUGUE_CONF_SQL_IGNORE_CASE, False)
    if fsql_dialect is None:
        fsql_dialect = get_current_conf().get(
            FUGUE_CONF_SQL_DIALECT, FUGUE_SQL_DEFAULT_DIALECT
        )
    dag = FugueSQLWorkflow(
        compile_conf={
            FUGUE_CONF_SQL_IGNORE_CASE: fsql_ignore_case,
            FUGUE_CONF_SQL_DIALECT: fsql_dialect,
        }
    )
    try:
        dag._sql(query, global_vars, local_vars, *args, **kwargs)
    except SyntaxError as ex:
        raise SyntaxError(str(ex)).with_traceback(None) from None
    return dag



from typing import Any

from triad.utils.assertion import assert_or_throw
from triad.utils.hash import to_uuid
from triad.utils.io import exists, join, makedirs, rm

from fugue.collections.partition import PartitionSpec
from fugue.collections.yielded import PhysicalYielded
from fugue.constants import FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH
from fugue.dataframe import DataFrame
from fugue.exceptions import FugueWorkflowCompileError, FugueWorkflowRuntimeError
from fugue.execution.execution_engine import ExecutionEngine


class Checkpoint(object):
    def __init__(
        self,
        deterministic: bool = False,
        permanent: bool = False,
        lazy: bool = False,
        **kwargs: Any,
    ):
        if deterministic:
            assert_or_throw(permanent, "deterministic checkpoint must be permanent")
        self.deterministic = deterministic
        self.permanent = permanent
        self.lazy = lazy
        self.kwargs = dict(kwargs)

    def run(self, df: DataFrame, path: "CheckpointPath") -> DataFrame:
        return df

    @property
    def is_null(self) -> bool:
        return True


class StrongCheckpoint(Checkpoint):
    def __init__(
        self,
        storage_type: str,
        obj_id: str,
        deterministic: bool,
        permanent: bool,
        lazy: bool = False,
        partition: Any = None,
        single: bool = False,
        namespace: Any = None,
        **save_kwargs: Any,
    ):
        super().__init__(
            deterministic=deterministic,
            permanent=permanent,
            lazy=lazy,
            fmt="",
            partition=PartitionSpec(partition),
            single=single,
            namespace=namespace,
            save_kwargs=dict(save_kwargs),
        )
        self._yield_func: Any = None
        self._obj_id = to_uuid(obj_id, namespace)
        self._yielded = PhysicalYielded(self._obj_id, storage_type=storage_type)

    def run(self, df: DataFrame, path: "CheckpointPath") -> DataFrame:
        if self._yielded.storage_type == "file":
            fpath = path.get_temp_file(self._obj_id, self.permanent)
            if not self.deterministic or not path.temp_file_exists(fpath):
                path.execution_engine.save_df(
                    df=df,
                    path=fpath,
                    format_hint=self.kwargs["fmt"],
                    mode="overwrite",
                    partition_spec=self.kwargs["partition"],
                    force_single=self.kwargs["single"],
                    **self.kwargs["save_kwargs"],
                )
            result = path.execution_engine.load_df(
                path=fpath, format_hint=self.kwargs["fmt"]
            )
            self._yielded.set_value(fpath)
        else:
            tb = path.get_table_name(self._obj_id, self.permanent)
            if (
                not self.deterministic
                or not path.execution_engine.sql_engine.table_exists(tb)
            ):
                path.execution_engine.sql_engine.save_table(
                    df=df,
                    table=tb,
                    partition_spec=self.kwargs["partition"],
                    **self.kwargs["save_kwargs"],
                )
            result = path.execution_engine.sql_engine.load_table(tb)
            self._yielded.set_value(tb)
        return result

    @property
    def yielded(self) -> PhysicalYielded:
        assert_or_throw(
            self.permanent,
            lambda: FugueWorkflowCompileError(f"yield is not allowed for {self}"),
        )
        return self._yielded

    @property
    def is_null(self) -> bool:
        return False


class WeakCheckpoint(Checkpoint):
    def __init__(
        self,
        lazy: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            deterministic=False,
            lazy=lazy,
            **kwargs,
        )

    def run(self, df: DataFrame, path: "CheckpointPath") -> DataFrame:
        return path.execution_engine.persist(df, lazy=self.lazy, **self.kwargs)

    @property
    def is_null(self) -> bool:
        return False


class CheckpointPath(object):
    def __init__(self, engine: ExecutionEngine):
        self._engine = engine
        self._log = engine.log
        self._path = engine.conf.get(FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH, "").strip()
        self._temp_path = ""

    @property
    def execution_engine(self) -> ExecutionEngine:
        return self._engine

    def init_temp_path(self, execution_id: str) -> str:
        if self._path == "":
            self._temp_path = ""
            return ""
        self._temp_path = join(self._path, execution_id)
        makedirs(self._temp_path, exist_ok=True)
        return self._temp_path

    def remove_temp_path(self):
        if self._temp_path != "":
            try:
                rm(self._temp_path, recursive=True)
            except Exception as e:  # pragma: no cover
                self._log.info("Unable to remove " + self._temp_path, e)

    def get_temp_file(self, obj_id: str, permanent: bool) -> str:
        path = self._path if permanent else self._temp_path
        assert_or_throw(
            path != "",
            FugueWorkflowRuntimeError(
                f"{FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH} is not set"
            ),
        )
        return join(path, obj_id + ".parquet")

    def get_table_name(self, obj_id: str, permanent: bool) -> str:
        path = self._path if permanent else self._temp_path
        return "temp_" + to_uuid(path, obj_id)[:5]

    def temp_file_exists(self, path: str) -> bool:
        try:
            return exists(path)
        except Exception:  # pragma: no cover
            return False



from typing import Any, Callable, Dict, Optional, Tuple

import ibis
from ibis.backends.pandas import Backend

from fugue import DataFrame, DataFrames, ExecutionEngine
from fugue.collections.sql import StructuredRawSQL, TempTableName
from fugue_ibis import IbisTable
from fugue_ibis._utils import to_ibis_schema
from fugue_ibis.execution.ibis_engine import IbisEngine, parse_ibis_engine

from .execution_engine import DuckDBEngine, DuckExecutionEngine


class DuckDBIbisEngine(IbisEngine):
    def select(
        self, dfs: DataFrames, ibis_func: Callable[[ibis.BaseBackend], IbisTable]
    ) -> DataFrame:
        be = _BackendWrapper().connect({})
        be.set_schemas(dfs)
        expr = ibis_func(be)
        sql = StructuredRawSQL.from_expr(
            str(
                ibis.postgres.compile(expr).compile(
                    compile_kwargs={"literal_binds": True}
                )
            ),
            prefix='"<tmpdf:',
            suffix='>"',
            dialect="postgres",
        )

        engine = DuckDBEngine(self.execution_engine)
        _dfs = DataFrames({be._name_map[k][0].key: v for k, v in dfs.items()})
        return engine.select(_dfs, sql)


@parse_ibis_engine.candidate(
    lambda obj, *args, **kwargs: isinstance(obj, DuckExecutionEngine)
    or (isinstance(obj, str) and obj in ["duck", "duckdb"])
)
def _to_duck_ibis_engine(obj: Any, engine: ExecutionEngine) -> Optional[IbisEngine]:
    return DuckDBIbisEngine(engine)


class _BackendWrapper(Backend):
    def set_schemas(self, dfs: DataFrames) -> None:
        self._schemas = {k: to_ibis_schema(v.schema) for k, v in dfs.items()}
        self._name_map: Dict[str, Tuple[TempTableName, IbisTable]] = {}

    def table(self, name: str, schema: Any = None) -> IbisTable:
        if name not in self._name_map:
            tn = TempTableName()
            tb = ibis.table(self._schemas[name], name=(str(tn)))
            self._name_map[name] = (tn, tb)
        return self._name_map[name][1]



# flake8: noqa
from ._utils import namespace_candidate
from .creator import Creator, creator, parse_creator, register_creator
from .outputter import Outputter, outputter, parse_outputter, register_outputter
from .processor import Processor, parse_processor, processor, register_processor
from .transformer import (
    CoTransformer,
    OutputCoTransformer,
    OutputTransformer,
    Transformer,
    cotransformer,
    output_cotransformer,
    output_transformer,
    parse_output_transformer,
    parse_transformer,
    register_output_transformer,
    register_transformer,
    transformer,
)



from abc import ABC, abstractmethod
from typing import Any

import pandas as pd
from triad import assert_or_throw

from fugue import DataFrames, Outputter
from fugue.exceptions import FugueWorkflowError


class Visualize(Outputter, ABC):
    def __init__(self, func: str) -> None:
        super().__init__()
        self._func = func

    def process(self, dfs: DataFrames) -> None:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        df = dfs[0].as_pandas()
        presort = self.partition_spec.presort
        presort_keys = list(presort.keys())
        presort_asc = list(presort.values())
        if len(presort_keys) > 0:
            df = df.sort_values(presort_keys, ascending=presort_asc).reset_index(
                drop=True
            )
        if len(self.partition_spec.partition_by) == 0:
            self._plot(df)
        else:
            keys: Any = (  # avoid pandas warning
                self.partition_spec.partition_by
                if len(self.partition_spec.partition_by) > 1
                else self.partition_spec.partition_by[0]
            )
            for _, gp in df.groupby(keys, dropna=False):
                self._plot(gp.reset_index(drop=True))

    @abstractmethod
    def _plot(self, df: pd.DataFrame) -> None:  # pragma: no cover
        raise NotImplementedError



import pytest


@pytest.fixture(scope="session")
def fugue_dask_client():
    from dask.distributed import Client
    import dask

    with Client(processes=True, n_workers=3, threads_per_worker=1) as client:
        dask.config.set({"dataframe.shuffle.method": "tasks"})
        dask.config.set({"dataframe.convert-string": False})
        yield client



from typing import Any, Optional

from .._utils.registry import fugue_plugin
from .dataset import AnyDataset, Dataset


@fugue_plugin
def as_fugue_dataset(data: AnyDataset, **kwargs: Any) -> Dataset:
    """Wrap the input as a :class:`~.Dataset`

    :param data: the dataset to be wrapped
    """
    if isinstance(data, Dataset) and len(kwargs) == 0:
        return data
    raise NotImplementedError(f"no registered dataset conversion for {type(data)}")


def show(
    data: AnyDataset, n: int = 10, with_count: bool = False, title: Optional[str] = None
) -> None:
    """Display the Dataset

    :param data: the dataset that can be recognized by Fugue
    :param n: number of rows to print, defaults to 10
    :param with_count: whether to show dataset count, defaults to False
    :param title: title of the dataset, defaults to None

    .. note::

        When ``with_count`` is True, it can trigger expensive calculation for
        a distributed dataframe. So if you call this function directly, you may
        need to :func:`fugue.execution.execution_engine.ExecutionEngine.persist`
        the dataset.
    """
    return as_fugue_dataset(data).show(n=n, with_count=with_count, title=title)


@fugue_plugin
def as_local(data: AnyDataset) -> AnyDataset:
    """Convert the dataset to a local dataset

    :param data: the dataset that can be recognized by Fugue
    """
    return as_local_bounded(data)


@fugue_plugin
def as_local_bounded(data: AnyDataset) -> AnyDataset:
    """Convert the dataset to a local bounded dataset

    :param data: the dataset that can be recognized by Fugue
    """
    raise NotImplementedError(
        f"no registered function to convert {type(data)} to a local bounded dataset"
    )


@fugue_plugin
def is_local(data: AnyDataset) -> bool:
    """Whether the dataset is local

    :param data: the dataset that can be recognized by Fugue
    """
    return as_fugue_dataset(data).is_local


@fugue_plugin
def is_bounded(data: AnyDataset) -> bool:
    """Whether the dataset is local

    :param data: the dataset that can be recognized by Fugue
    """
    return as_fugue_dataset(data).is_bounded


@fugue_plugin
def is_empty(data: AnyDataset) -> bool:
    """Whether the dataset is empty

    :param data: the dataset that can be recognized by Fugue
    """
    return as_fugue_dataset(data).empty


@fugue_plugin
def count(data: AnyDataset) -> int:
    """The number of elements in the dataset

    :param data: the dataset that can be recognized by Fugue
    """
    return as_fugue_dataset(data).count()


@fugue_plugin
def get_num_partitions(data: AnyDataset) -> bool:
    """Get the number of partitions of the dataset

    :param data: the dataset that can be recognized by Fugue
    """
    return as_fugue_dataset(data).num_partitions



import os

from setuptools import find_packages, setup

from fugue_version import __version__

with open("README.md") as f:
    _text = ["# Fugue"] + f.read().splitlines()[1:]
    LONG_DESCRIPTION = "\n".join(_text)


def get_version() -> str:
    tag = os.environ.get("RELEASE_TAG", "")
    if "dev" in tag.split(".")[-1]:
        return tag
    if tag != "":
        assert tag == __version__, "release tag and version mismatch"
    return __version__


setup(
    name="fugue",
    version=get_version(),
    packages=find_packages(include=["fugue*"]),
    description="An abstraction layer for distributed computation",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    author="The Fugue Development Team",
    author_email="hello@fugue.ai",
    keywords="distributed spark dask ray sql dsl domain specific language",
    url="http://github.com/fugue-project/fugue",
    install_requires=[
        "triad>=0.9.3",
        "adagio>=0.2.4",
        # sql dependencies
        "qpd>=0.4.4",
        "fugue-sql-antlr>=0.1.6",
        "sqlglot",
        "jinja2",
    ],
    extras_require={
        "sql": [
            "qpd>=0.4.4",
            "fugue-sql-antlr>=0.1.6",
            "sqlglot",
            "jinja2",
        ],
        "cpp_sql_parser": ["fugue-sql-antlr[cpp]>=0.1.6"],
        "spark": ["pyspark>=3.1.1"],
        "dask": [
            "dask[distributed,dataframe]>=2023.5.0",
            "pyarrow>=7.0.0",
            "pandas>=2.0.2",
        ],
        "ray": ["ray[data]>=2.4.0", "duckdb>=0.5.0", "pyarrow>=6.0.1"],
        "duckdb": [
            "duckdb>=0.5.0",
            "numpy",
        ],
        "polars": ["polars"],
        "ibis": ["ibis-framework>=3.2.0,<6"],
        "notebook": ["notebook", "jupyterlab", "ipython>=7.10.0"],
        "all": [
            "sqlglot",
            "jinja2",
            "fugue-sql-antlr[cpp]>=0.1.6",
            "pyspark>=3.1.1",
            "dask[distributed,dataframe]>=2023.5.0",
            "dask-sql",
            "ray[data]>=2.4.0",
            "notebook",
            "jupyterlab",
            "ipython>=7.10.0",
            "duckdb>=0.5.0",
            "pyarrow>=6.0.1",
            "pandas>=2.0.2",
            "ibis-framework>=3.2.0,<6",
            "polars",
        ],
    },
    classifiers=[
        # "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    python_requires=">=3.8",
    package_data={"fugue": ["py.typed"], "fugue_notebook": ["nbextension/*"]},
    entry_points={
        "fugue.plugins": [
            "ibis = fugue_ibis[ibis]",
            "duckdb = fugue_duckdb.registry[duckdb]",
            "duckdb_ibis = fugue_duckdb.ibis_engine[duckdb,ibis]",
            "spark = fugue_spark.registry[spark]",
            "spark_ibis = fugue_spark.ibis_engine[spark,ibis]",
            "dask = fugue_dask.registry[dask]",
            "dask_ibis = fugue_dask.ibis_engine[dask,ibis]",
            "ray = fugue_ray.registry[ray]",
            "polars = fugue_polars.registry[polars]",
        ],
        "pytest11": [
            "fugue_test_dask = fugue_test.plugins.dask[dask]",
            "fugue_test_ray = fugue_test.plugins.ray[ray]",
            "fugue_test_duckdb = fugue_test.plugins.duckdb[duckdb]",
            "fugue_test_misc = fugue_test.plugins.misc",
        ],
    },
)



from typing import Any, Callable, List, Optional, Type, Union

import pandas as pd
from triad import ParamDict, assert_or_throw
from triad.utils.convert import to_instance

from .._utils.registry import fugue_plugin
from ..exceptions import FuguePluginsRegistrationError
from .execution_engine import (
    _FUGUE_EXECUTION_ENGINE_CONTEXT,
    _FUGUE_GLOBAL_EXECUTION_ENGINE_CONTEXT,
    ExecutionEngine,
    SQLEngine,
)
from .native_execution_engine import NativeExecutionEngine


def register_execution_engine(
    name_or_type: Union[str, Type], func: Callable, on_dup="overwrite"
) -> None:
    """Register :class:`~fugue.execution.execution_engine.ExecutionEngine` with
    a given name.

    :param name_or_type: alias of the execution engine, or type of an object that
      can be converted to an execution engine
    :param func: a callable taking |ParamsLikeObject| and ``**kwargs`` and returning an
      :class:`~fugue.execution.execution_engine.ExecutionEngine` instance
    :param on_dup: action on duplicated ``name``. It can be "overwrite", "ignore"
      (not overwriting), defaults to "overwrite".

    .. admonition:: Examples

        Alias registration examples:

        .. code-block:: python

            # create a new engine with name my (overwrites if existed)
            register_execution_engine("my", lambda conf: MyExecutionEngine(conf))

            # 0
            make_execution_engine("my")
            make_execution_engine("my", {"myconfig":"value})

            # 1
            dag = FugueWorkflow()
            dag.create([[0]],"a:int").show()
            dag.run("my", {"myconfig":"value})

            # 2
            fsql('''
            CREATE [[0]] SCHEMA a:int
            PRINT
            ''').run("my")

        Type registration examples:

        .. code-block:: python

            from pyspark.sql import SparkSession
            from fugue_spark import SparkExecutionEngine
            from fugue import fsql

            register_execution_engine(
                SparkSession,
                lambda session, conf: SparkExecutionEngine(session, conf))

            spark_session = SparkSession.builder.getOrCreate()

            fsql('''
            CREATE [[0]] SCHEMA a:int
            PRINT
            ''').run(spark_session)
    """
    if isinstance(name_or_type, str):
        nm = name_or_type
        parse_execution_engine.register(  # type: ignore
            func=lambda engine, conf, **kwargs: func(conf, **kwargs),
            matcher=lambda engine, conf, **kwargs: isinstance(engine, str)
            and engine == nm,
            priority=_get_priority(on_dup),
        )
    else:  # type
        tp = name_or_type
        parse_execution_engine.register(  # type: ignore
            func=lambda engine, conf, **kwargs: func(engine, conf, **kwargs),
            matcher=lambda engine, conf, **kwargs: isinstance(engine, tp),
            priority=_get_priority(on_dup),
        )


def register_default_execution_engine(func: Callable, on_dup="overwrite") -> None:
    """Register :class:`~fugue.execution.execution_engine.ExecutionEngine` as the
    default engine.

    :param func: a callable taking |ParamsLikeObject| and ``**kwargs`` and returning an
      :class:`~fugue.execution.execution_engine.ExecutionEngine` instance
    :param on_dup: action on duplicated ``name``. It can be "overwrite", "ignore"
      (not overwriting), defaults to "overwrite".

    .. admonition:: Examples

        .. code-block:: python

            # create a new engine with name my (overwrites if existed)
            register_default_execution_engine(lambda conf: MyExecutionEngine(conf))

            # the following examples will use MyExecutionEngine

            # 0
            make_execution_engine()
            make_execution_engine(None, {"myconfig":"value})

            # 1
            dag = FugueWorkflow()
            dag.create([[0]],"a:int").show()
            dag.run(None, {"myconfig":"value})

            # 2
            fsql('''
            CREATE [[0]] SCHEMA a:int
            PRINT
            ''').run("", {"myconfig":"value})
    """
    parse_execution_engine.register(  # type: ignore
        func=lambda engine, conf, **kwargs: func(conf, **kwargs),
        matcher=lambda engine, conf, **kwargs: engine is None
        or (isinstance(engine, str) and engine == ""),
        priority=_get_priority(on_dup),
    )


def register_sql_engine(name: str, func: Callable, on_dup="overwrite") -> None:
    """Register :class:`~fugue.execution.execution_engine.SQLEngine` with
    a given name.

    :param name: name of the SQL engine
    :param func: a callable taking
      :class:`~fugue.execution.execution_engine.ExecutionEngine`
      and ``**kwargs`` and returning a
      :class:`~fugue.execution.execution_engine.SQLEngine` instance
    :param on_dup: action on duplicated ``name``. It can be "overwrite", "ignore"
      (not overwriting), defaults to "overwrite".

    .. admonition:: Examples

        .. code-block:: python

            # create a new engine with name my (overwrites if existed)
            register_sql_engine("mysql", lambda engine: MySQLEngine(engine))

            # create execution engine with MySQLEngine as the default
            make_execution_engine(("", "mysql"))

            # create DaskExecutionEngine with MySQLEngine as the default
            make_execution_engine(("dask", "mysql"))

            # default execution engine + MySQLEngine
            with FugueWorkflow() as dag:
                dag.create([[0]],"a:int").show()
            dag.run(("","mysql"))
    """
    nm = name
    parse_sql_engine.register(  # type: ignore
        func=lambda engine, execution_engine, **kwargs: func(
            execution_engine, **kwargs
        ),
        matcher=lambda engine, execution_engine, **kwargs: isinstance(engine, str)
        and engine == nm,
        priority=_get_priority(on_dup),
    )


def register_default_sql_engine(func: Callable, on_dup="overwrite") -> None:
    """Register :class:`~fugue.execution.execution_engine.SQLEngine` as the
    default engine

    :param func: a callable taking
      :class:`~fugue.execution.execution_engine.ExecutionEngine`
      and ``**kwargs`` and returning a
      :class:`~fugue.execution.execution_engine.SQLEngine` instance
    :param on_dup: action on duplicated ``name``. It can be "overwrite", "ignore"
      (not overwriting) or "throw" (throw exception), defaults to "overwrite".

    :raises KeyError: if ``on_dup`` is ``throw`` and the ``name`` already exists

    .. note::

        You should be careful to use this function, because when you set a custom
        SQL engine as default, all execution engines you create will use this SQL
        engine unless you are explicit. For example if you set the default SQL engine
        to be a Spark specific one, then if you start a NativeExecutionEngine, it will
        try to use it and will throw exceptions.

        So it's always a better idea to use ``register_sql_engine`` instead

    .. admonition:: Examples

        .. code-block:: python

            # create a new engine with name my (overwrites if existed)
            register_default_sql_engine(lambda engine: MySQLEngine(engine))

            # create NativeExecutionEngine with MySQLEngine as the default
            make_execution_engine()

            # create SparkExecutionEngine with MySQLEngine instead of SparkSQLEngine
            make_execution_engine("spark")

            # NativeExecutionEngine with MySQLEngine
            with FugueWorkflow() as dag:
                dag.create([[0]],"a:int").show()
            dag.run()
    """
    parse_sql_engine.register(  # type: ignore
        func=lambda engine, execution_engine, **kwargs: func(
            execution_engine, **kwargs
        ),
        matcher=lambda engine, execution_engine, **kwargs: engine is None
        or (isinstance(engine, str) and engine == ""),
        priority=_get_priority(on_dup),
    )


def try_get_context_execution_engine() -> Optional[ExecutionEngine]:
    """If the global execution engine is set (see
    :meth:`~fugue.execution.execution_engine.ExecutionEngine.set_global`)
    or the context is set (see
    :meth:`~fugue.execution.execution_engine.ExecutionEngine.as_context`),
    then return the engine, else return None
    """
    engine = _FUGUE_EXECUTION_ENGINE_CONTEXT.get()
    if engine is None:
        engine = _FUGUE_GLOBAL_EXECUTION_ENGINE_CONTEXT.get()
    return engine


def make_execution_engine(
    engine: Any = None,
    conf: Any = None,
    infer_by: Optional[List[Any]] = None,
    **kwargs: Any,
) -> ExecutionEngine:
    """Create :class:`~fugue.execution.execution_engine.ExecutionEngine`
    with specified ``engine``

    :param engine: it can be empty string or null (use the default execution
      engine), a string (use the registered execution engine), an
      :class:`~fugue.execution.execution_engine.ExecutionEngine` type, or
      the :class:`~fugue.execution.execution_engine.ExecutionEngine` instance
      , or a tuple of two values where the first value represents execution
      engine and the second value represents the sql engine (you can use ``None``
      for either of them to use the default one), defaults to None
    :param conf: |ParamsLikeObject|, defaults to None
    :param infer_by: List of objects that can be used to infer the execution
      engine using :func:`~.infer_execution_engine`
    :param kwargs: additional parameters to initialize the execution engine

    :return: the :class:`~fugue.execution.execution_engine.ExecutionEngine`
      instance

    .. note::

        This function finds/constructs the engine in the following order:

        * If ``engine`` is None, it first try to see if there is any defined
          context engine to use (=> engine)
        * If ``engine`` is still empty, then it will try to get the global execution
          engine. See
          :meth:`~fugue.execution.execution_engine.ExecutionEngine.set_global`
        * If ``engine`` is still empty, then if ``infer_by``
          is given, it will try to infer the execution engine (=> engine)
        * If ``engine`` is still empty, then it will construct the default
          engine defined by :func:`~.register_default_execution_engine`  (=> engine)
        * Now, ``engine`` must not be empty, if it is an object other than
          :class:`~fugue.execution.execution_engine.ExecutionEngine`, we will use
          :func:`~.parse_execution_engine` to construct (=> engine)
        * Now, ``engine`` must have been an ExecutionEngine object. We update its
          SQL engine if specified, then update its config using ``conf`` and ``kwargs``

    .. admonition:: Examples

        .. code-block:: python

            register_default_execution_engine(lambda conf: E1(conf))
            register_execution_engine("e2", lambda conf, **kwargs: E2(conf, **kwargs))

            register_sql_engine("s", lambda conf: S2(conf))

            # E1 + E1.create_default_sql_engine()
            make_execution_engine()

            # E2 + E2.create_default_sql_engine()
            make_execution_engine(e2)

            # E1 + S2
            make_execution_engine((None, "s"))

            # E2(conf, a=1, b=2) + S2
            make_execution_engine(("e2", "s"), conf, a=1, b=2)

            # SparkExecutionEngine + SparkSQLEngine
            make_execution_engine(SparkExecutionEngine)
            make_execution_engine(SparkExecutionEngine(spark_session, conf))

            # SparkExecutionEngine + S2
            make_execution_engine((SparkExecutionEngine, "s"))

            # assume object e2_df can infer E2 engine
            make_execution_engine(infer_by=[e2_df])  # an E2 engine

            # global
            e_global = E1(conf)
            e_global.set_global()
            make_execution_engine()  # e_global

            # context
            with E2(conf).as_context() as ec:
                make_execution_engine()  # ec
            make_execution_engine()  # e_global
    """
    if engine is None:
        engine = try_get_context_execution_engine()
        if engine is None and infer_by is not None:
            engine = infer_execution_engine(infer_by)

    if isinstance(engine, tuple):
        execution_engine = make_execution_engine(engine[0], conf=conf, **kwargs)
        sql_engine = make_sql_engine(engine[1], execution_engine)
        execution_engine.set_sql_engine(sql_engine)
        return execution_engine
    if isinstance(engine, ExecutionEngine):
        result = engine
    else:
        result = parse_execution_engine(engine, conf, **kwargs)
        sql_engine = make_sql_engine(None, result)
        result.set_sql_engine(sql_engine)
    result.conf.update(conf, on_dup=ParamDict.OVERWRITE)
    result.conf.update(kwargs, on_dup=ParamDict.OVERWRITE)
    return result


@fugue_plugin
def parse_execution_engine(
    engine: Any = None, conf: Any = None, **kwargs: Any
) -> ExecutionEngine:
    """Parse object as
    :class:`~fugue.execution.execution_engine.ExecutionEngine`

    :param engine: it can be empty string or null (use the default execution
      engine), a string (use the registered execution engine), an
      :class:`~fugue.execution.execution_engine.ExecutionEngine` type, or
      the :class:`~fugue.execution.execution_engine.ExecutionEngine` instance
      , or a tuple of two values where the first value represents execution
      engine and the second value represents the sql engine (you can use ``None``
      for either of them to use the default one), defaults to None
    :param conf: |ParamsLikeObject|, defaults to None
    :param kwargs: additional parameters to initialize the execution engine

    :return: the :class:`~fugue.execution.execution_engine.ExecutionEngine`
      instance

    .. admonition:: Examples

        .. code-block:: python

            register_default_execution_engine(lambda conf: E1(conf))
            register_execution_engine("e2", lambda conf, **kwargs: E2(conf, **kwargs))

            register_sql_engine("s", lambda conf: S2(conf))

            # E1 + E1.default_sql_engine
            make_execution_engine()

            # E2 + E2.default_sql_engine
            make_execution_engine(e2)

            # E1 + S2
            make_execution_engine((None, "s"))

            # E2(conf, a=1, b=2) + S2
            make_execution_engine(("e2", "s"), conf, a=1, b=2)

            # SparkExecutionEngine + SparkSQLEngine
            make_execution_engine(SparkExecutionEngine)
            make_execution_engine(SparkExecutionEngine(spark_session, conf))

            # SparkExecutionEngine + S2
            make_execution_engine((SparkExecutionEngine, "s"))
    """
    if engine is None or (isinstance(engine, str) and engine == ""):
        return NativeExecutionEngine(conf)
    try:
        return to_instance(engine, ExecutionEngine, kwargs=dict(conf=conf, **kwargs))
    except Exception as e:
        raise FuguePluginsRegistrationError(
            f"Fugue execution engine is not recognized ({engine}, {conf}, {kwargs})."
            " You may need to register a parser for it."
        ) from e


def is_pandas_or(objs: List[Any], obj_type: Any) -> bool:
    """Check whether the input contains at least one ``obj_type`` object and the
    rest are Pandas DataFrames. This function is a utility function for extending
    :func:`~.infer_execution_engine`

    :param objs: the list of objects to check
    :return: whether all objs are of type ``obj_type`` or pandas DataFrame and at
      least one is of type ``obj_type``
    """
    tc = 0
    for obj in objs:
        if not isinstance(obj, pd.DataFrame):
            if isinstance(obj, obj_type):
                tc += 1
            else:
                return False
    return tc > 0


@fugue_plugin
def infer_execution_engine(obj: List[Any]) -> Any:
    """Infer the correspondent ExecutionEngine based on the input objects. This is
    used in express functions.

    :param objs: the objects
    :return: if the inference succeeded, it returns an object that can be used by
      :func:`~.parse_execution_engine` in the ``engine`` field to construct an
      ExecutionEngine. Otherwise, it returns None.

    .. admonition:: Examples

        .. code-block:: python

            from fugue import transform

            transform(spark_df)

        In this example, the SparkExecutionEngine is inferred from spark_df, it
        is equivalent to:

        .. code-block:: python

            from fugue import transform

            transform(spark_df, engine=current_spark_session)
    """
    return None


def make_sql_engine(
    engine: Any = None,
    execution_engine: Optional[ExecutionEngine] = None,
    **kwargs: Any,
) -> SQLEngine:
    """Create :class:`~fugue.execution.execution_engine.SQLEngine`
    with specified ``engine``

    :param engine: it can be empty string or null (use the default SQL
      engine), a string (use the registered SQL engine), an
      :class:`~fugue.execution.execution_engine.SQLEngine` type, or
      the :class:`~fugue.execution.execution_engine.SQLEngine` instance
      (you can use ``None`` to use the default one), defaults to None
    :param execution_engine: the
      :class:`~fugue.execution.execution_engine.ExecutionEngine` instance
      to create
      the :class:`~fugue.execution.execution_engine.SQLEngine`. Normally you
      should always provide this value.
    :param kwargs: additional parameters to initialize the sql engine

    :return: the :class:`~fugue.execution.execution_engine.SQLEngine`
      instance

    .. note::

        For users, you normally don't need to call this function directly.
        Use ``make_execution_engine`` instead

    .. admonition:: Examples

        .. code-block:: python

            register_default_sql_engine(lambda conf: S1(conf))
            register_sql_engine("s2", lambda conf: S2(conf))

            engine = NativeExecutionEngine()

            # S1(engine)
            make_sql_engine(None, engine)

            # S1(engine, a=1)
            make_sql_engine(None, engine, a=1)

            # S2(engine)
            make_sql_engine("s2", engine)
    """
    if isinstance(engine, SQLEngine):
        assert_or_throw(
            execution_engine is None and len(kwargs) == 0,
            lambda: ValueError(
                f"{engine} is an instance, can't take arguments "
                f"execution_engine={execution_engine}, kwargs={kwargs}"
            ),
        )
        return engine
    return parse_sql_engine(engine, execution_engine, **kwargs)


@fugue_plugin
def parse_sql_engine(
    engine: Any = None,
    execution_engine: Optional[ExecutionEngine] = None,
    **kwargs: Any,
) -> SQLEngine:
    """Create :class:`~fugue.execution.execution_engine.SQLEngine`
    with specified ``engine``

    :param engine: it can be empty string or null (use the default SQL
      engine), a string (use the registered SQL engine), an
      :class:`~fugue.execution.execution_engine.SQLEngine` type, or
      the :class:`~fugue.execution.execution_engine.SQLEngine` instance
      (you can use ``None`` to use the default one), defaults to None
    :param execution_engine: the
      :class:`~fugue.execution.execution_engine.ExecutionEngine` instance
      to create
      the :class:`~fugue.execution.execution_engine.SQLEngine`. Normally you
      should always provide this value.
    :param kwargs: additional parameters to initialize the sql engine

    :return: the :class:`~fugue.execution.execution_engine.SQLEngine`
      instance

    .. note::

        For users, you normally don't need to call this function directly.
        Use ``make_execution_engine`` instead

    .. admonition:: Examples

        .. code-block:: python

            register_default_sql_engine(lambda conf: S1(conf))
            register_sql_engine("s2", lambda conf: S2(conf))

            engine = NativeExecutionEngine()

            # S1(engine)
            make_sql_engine(None, engine)

            # S1(engine, a=1)
            make_sql_engine(None, engine, a=1)

            # S2(engine)
            make_sql_engine("s2", engine)
    """
    if engine is None or (isinstance(engine, str) and engine == ""):
        assert_or_throw(
            execution_engine is not None,
            ValueError("execution_engine must be provided"),
        )
        return execution_engine.sql_engine  # type: ignore
    try:
        return to_instance(
            engine, SQLEngine, kwargs=dict(execution_engine=execution_engine, **kwargs)
        )
    except Exception as e:
        raise FuguePluginsRegistrationError(
            f"Fugue SQL engine is not recognized ({engine}, {kwargs})."
            " You may need to register a parser for it."
        ) from e


def _get_priority(on_dup: str) -> int:
    if on_dup == "overwrite":
        return 2
    if on_dup == "ignore":
        return 0
    raise ValueError(f"{on_dup} is not valid")



import inspect
from typing import Callable, Optional

from triad.utils.assertion import assert_or_throw

_COMMENT_SCHEMA_ANNOTATION = "schema"


def parse_comment_annotation(func: Callable, annotation: str) -> Optional[str]:
    """Parse comment annotation above the function. It try to find
    comment lines starts with the annotation from bottom up, and will use the first
    occurrance as the result.

    :param func: the function
    :param annotation: the annotation string
    :return: schema hint string

    .. admonition:: Examples

        .. code-block:: python

            # schema: a:int,b:str
            #schema:a:int,b:int # more comment
            # some comment
            def dummy():
                pass

            assert "a:int,b:int" == parse_comment_annotation(dummy, "schema:")
    """
    for orig in reversed((inspect.getcomments(func) or "").splitlines()):
        start = orig.find(":")
        if start <= 0:
            continue
        actual = orig[:start].replace("#", "", 1).strip()
        if actual != annotation:
            continue
        end = orig.find("#", start)
        s = orig[start + 1 : (end if end > 0 else len(orig))].strip()
        return s
    return None


def parse_output_schema_from_comment(func: Callable) -> Optional[str]:
    """Parse schema hint from the comments above the function. It try to find
    comment lines starts with `schema:` from bottom up, and will use the first
    occurrance as the hint.

    :param func: the function
    :return: schema hint string

    .. admonition:: Examples

        .. code-block:: python

            # schema: a:int,b:str
            #schema:a:int,b:int # more comment
            # some comment
            def dummy():
                pass

            assert "a:int,b:int" == parse_output_schema_from_comment(dummy)
    """
    res = parse_comment_annotation(func, _COMMENT_SCHEMA_ANNOTATION)
    if res is None:
        return None
    assert_or_throw(res != "", SyntaxError("incorrect schema annotation"))
    return res.strip()


def is_class_method(func: Callable) -> bool:
    sig = inspect.signature(func)
    # TODO: this is not the best way
    return "self" in sig.parameters
